"""
Public API for managing images in a Splitgraph repository.
"""

import itertools
import logging
import re
from contextlib import contextmanager
from datetime import datetime
from io import TextIOWrapper
from random import getrandbits
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from psycopg2.sql import SQL, Composed, Identifier

from splitgraph.config import (
    CONFIG,
    FDW_CLASS,
    SPLITGRAPH_API_SCHEMA,
    SPLITGRAPH_META_SCHEMA,
    get_singleton,
)
from splitgraph.core.fragment_manager import ExtraIndexInfo
from splitgraph.core.image import Image
from splitgraph.core.image_manager import ImageManager
from splitgraph.core.sql import insert, select, validate_import_sql
from splitgraph.core.table import Table
from splitgraph.core.types import TableSchema
from splitgraph.engine.postgres.engine import PostgresEngine
from splitgraph.exceptions import (
    CheckoutError,
    IncompleteObjectUploadError,
    RepositoryNotFoundError,
    TableNotFoundError,
)

from ..engine import ResultShape
from .common import (
    aggregate_changes,
    gather_sync_metadata,
    get_temporary_table_id,
    manage_audit,
    manage_audit_triggers,
    set_head,
    set_tags_batch,
    slow_diff,
    unmount_schema,
)
from .engine import get_engine, lookup_repository
from .object_manager import ObjectManager
from .output import pluralise


class Repository:
    """
    Splitgraph repository API
    """

    # Whilst these are enforced by PostgreSQL in the Splitgraph schema, it's good to
    # prevalidate namespace/repository values when the Repository object is constructed.
    _MAX_NAMESPACE_LEN = 64
    _MAX_REPOSITORY_LEN = 64
    _NAMESPACE_RE = re.compile(r"^[-A-Za-z0-9_]*$")
    _REPOSITORY_RE = re.compile(r"^[-A-Za-z0-9_]+$")

    def __init__(
        self,
        namespace: str,
        repository: str,
        engine: Optional[PostgresEngine] = None,
        object_engine: Optional[PostgresEngine] = None,
        object_manager: Optional[ObjectManager] = None,
    ) -> None:
        if len(namespace) > self._MAX_NAMESPACE_LEN or not self._NAMESPACE_RE.match(namespace):
            raise ValueError(
                f"Invalid namespace '{namespace}'. Namespace must contain at most 64 "
                "alphanumerics, dashes or underscores."
            )

        if len(repository) > self._MAX_REPOSITORY_LEN or not self._REPOSITORY_RE.match(repository):
            raise ValueError(
                f"Invalid repository '{repository}'. Repository must contain at most 64 "
                "alphanumerics, dashes or underscores."
            )

        self.namespace = namespace
        self.repository = repository

        self.engine = engine or get_engine()
        # Add an option to use a different engine class for storing cached table fragments (e.g. a different
        # PostgreSQL connection or even a different database engine altogether).
        self.object_engine = object_engine or self.engine
        self.images = ImageManager(self)
        """A :class:`splitgraph.core.image.ImageManager` instance that performs operations 
        (checkout, delete etc) on this repository's images.
        """

        # consider making this an injected/a singleton for a given engine
        # since it's global for the whole engine but calls (e.g. repo.objects.cleanup()) make it
        # look like it's the manager for objects related to a repo.
        self.objects = object_manager or ObjectManager(
            object_engine=self.object_engine, metadata_engine=self.engine
        )
        """A :class:`splitgraph.core.object_manager.ObjectManager` instance that performs operations on
        the objects on this repository's engine (not just objects belonging to this repository).
        """

    @classmethod
    def from_template(
        cls,
        template: "Repository",
        namespace: Optional[str] = None,
        repository: Optional[str] = None,
        engine: Optional[PostgresEngine] = None,
        object_engine: Optional[PostgresEngine] = None,
    ) -> "Repository":
        """Create a Repository from an existing one replacing some of its attributes."""
        # If engine has been overridden but not object_engine, also override the object_engine (maintain
        # the intended behaviour of overriding engine repointing the whole repository)
        return cls(
            namespace or template.namespace,
            repository or template.repository,
            engine or template.engine,
            object_engine or engine or template.object_engine,
        )

    @classmethod
    def from_schema(cls, schema: str) -> "Repository":
        """Convert a Postgres schema name of the format `namespace/repository` to a Splitgraph repository object."""
        if "/" in schema:
            namespace, repository = schema.split("/")
            return cls(namespace, repository)
        return cls("", schema)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Repository):
            return NotImplemented
        return self.namespace == other.namespace and self.repository == other.repository

    def to_schema(self) -> str:
        """Returns the engine schema that this repository gets checked out into."""
        return self.namespace + "/" + self.repository if self.namespace else self.repository

    def __repr__(self) -> str:
        repr = "Repository %s on %s" % (self.to_schema(), self.engine.name)
        if self.engine != self.object_engine:
            repr += " (object engine %s)" % self.object_engine.name
        return repr

    __str__ = to_schema

    def __hash__(self) -> int:
        return hash(self.namespace) * hash(self.repository)

    # --- GENERAL REPOSITORY MANAGEMENT ---

    def rollback_engines(self) -> None:
        """
        Rollback the underlying transactions on both engines that the repository uses.
        """
        self.engine.rollback()
        if self.engine != self.object_engine:
            self.object_engine.rollback()

    def commit_engines(self) -> None:
        """
        Commit the underlying transactions on both engines that the repository uses.
        """
        self.engine.commit()
        if self.engine != self.object_engine:
            self.object_engine.commit()

    @manage_audit
    def init(self) -> None:
        """
        Initializes an empty repo with an initial commit (hash 0000...)
        """
        self.object_engine.create_schema(self.to_schema())
        initial_image = "0" * 64
        self.engine.run_sql(
            insert("images", ("image_hash", "namespace", "repository", "parent_id", "created")),
            (initial_image, self.namespace, self.repository, None, datetime.utcnow()),
        )
        # Strictly speaking this is redundant since the checkout (of the "HEAD" commit) updates the tag table.
        self.engine.run_sql(
            insert("tags", ("namespace", "repository", "image_hash", "tag")),
            (self.namespace, self.repository, initial_image, "HEAD"),
        )

    def delete(self, unregister: bool = True, uncheckout: bool = True) -> None:
        """
        Discards all changes to a given repository and optionally all of its history,
        as well as deleting the Postgres schema that it might be checked out into.
        Doesn't delete any cached physical objects.

        After performing this operation, this object becomes invalid and must be discarded,
        unless init() is called again.

        :param unregister: Whether to purge repository history/metadata
        :param uncheckout: Whether to delete the actual checked out repo. This has no effect
            if the repository is backed by a registry (rather than a local engine).
        """
        # Make sure to discard changes to this repository if they exist, otherwise they might
        # be applied/recorded if a new repository with the same name appears.
        if uncheckout and not self.object_engine.registry:
            self.object_engine.discard_pending_changes(self.to_schema())

            # Dispose of the foreign servers (LQ FDW, other FDWs) for this schema if it exists
            # (otherwise its connection won't be recycled and we can get deadlocked).
            unmount_schema(self.object_engine, self.to_schema())
        if unregister:
            # Use the API call in case we're deleting a remote repository.
            self.engine.run_sql(
                SQL("SELECT {}.delete_repository(%s,%s)").format(Identifier(SPLITGRAPH_API_SCHEMA)),
                (self.namespace, self.repository),
            )

            # On local repos, also forget about the repository's upstream.
            if not self.object_engine.registry:
                self.object_engine.run_sql(
                    SQL("DELETE FROM {}.{} WHERE namespace = %s AND repository = %s").format(
                        Identifier(SPLITGRAPH_META_SCHEMA), Identifier("upstream")
                    ),
                    (self.namespace, self.repository),
                )
        self.engine.commit()

    @property
    def upstream(self):
        """
        The remote upstream repository that this local repository tracks.
        """
        result = self.engine.run_sql(
            select(
                "upstream",
                "remote_name, remote_namespace, remote_repository",
                "namespace = %s AND repository = %s",
            ),
            (self.namespace, self.repository),
            return_shape=ResultShape.ONE_MANY,
        )
        if result is None:
            return result
        try:
            engine = get_engine(result[0])
        # Ignore errors if the engine doesn't exist or the params aren't in the config
        except (KeyError, ValueError):
            logging.warning(
                "Repository %s/%s has upstream on remote %s which doesn't exist in the config.",
                self.namespace,
                self.repository,
                result[0],
            )
            return None

        return Repository(namespace=result[1], repository=result[2], engine=engine)

    @upstream.setter
    def upstream(self, remote_repository: "Repository"):
        """
        Sets the upstream remote + repository that this repository tracks.

        :param remote_repository: Remote Repository object
        """
        self.engine.run_sql(
            SQL(
                "INSERT INTO {0}.upstream (namespace, repository, "
                "remote_name, remote_namespace, remote_repository) VALUES (%s, %s, %s, %s, %s)"
                " ON CONFLICT (namespace, repository) DO UPDATE SET "
                "remote_name = excluded.remote_name, remote_namespace = excluded.remote_namespace, "
                "remote_repository = excluded.remote_repository WHERE "
                "upstream.namespace = excluded.namespace "
                "AND upstream.repository = excluded.repository"
            ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
            (
                self.namespace,
                self.repository,
                remote_repository.engine.name,
                remote_repository.namespace,
                remote_repository.repository,
            ),
        )

    @upstream.deleter
    def upstream(self):
        """
        Deletes the upstream remote + repository for a local repository.
        """
        self.engine.run_sql(
            SQL("DELETE FROM {0}.upstream WHERE namespace = %s AND repository = %s").format(
                Identifier(SPLITGRAPH_META_SCHEMA)
            ),
            (self.namespace, self.repository),
            return_shape=None,
        )

    def get_size(self) -> int:
        """
        Get the physical size used by the repository's data, counting objects that are used
        by multiple images only once. This is calculated from the metadata, the on-disk
        footprint might be smaller if not all of repository's objects have been downloaded.

        :return: Size of the repository in bytes.
        """
        return cast(
            int,
            self.engine.run_sql(
                select("get_repository_size", table_args="(%s,%s)", schema=SPLITGRAPH_API_SCHEMA),
                (self.namespace, self.repository),
                return_shape=ResultShape.ONE_ONE,
            )
            or 0,
        )

    def get_local_size(self) -> int:
        """
        Get the actual size used by this repository's downloaded objects.

        This might still be double-counted if the repository shares objects
        with other repositores.

        :return: Size of the repository in bytes.
        """
        repo_objects = self.objects.get_objects_for_repository(self)
        local_objects = self.objects.get_downloaded_objects(limit_to=repo_objects)
        local_object_meta = self.objects.get_object_meta(local_objects)
        return sum(o.size for o in local_object_meta.values())

    # --- COMMITS / CHECKOUTS ---

    @contextmanager
    def materialized_table(
        self, table_name: str, image_hash: Optional[str]
    ) -> Iterator[Tuple[str, str]]:
        """A context manager that returns a pointer to a read-only materialized table in a given image.
        The table is deleted on exit from the context manager.

        :param table_name: Name of the table
        :param image_hash: Image hash to materialize
        :return: (schema, table_name) where the materialized table is located.
        """
        if image_hash is None:
            # No image hash -- just return the current staging table.
            yield self.to_schema(), table_name
            return  # make sure we don't fall through after the user is finished

        table = self.images.by_hash(image_hash).get_table(table_name)
        # Materialize the table even if it's a single object to discard the upsert-delete flag.
        new_id = get_temporary_table_id()
        table.materialize(new_id, destination_schema=SPLITGRAPH_META_SCHEMA)
        try:
            yield SPLITGRAPH_META_SCHEMA, new_id
        finally:
            # Maybe some cache management/expiry strategies here
            self.object_engine.delete_table(SPLITGRAPH_META_SCHEMA, new_id)
            self.object_engine.commit()

    @property
    def head_strict(self) -> Image:
        """Return the HEAD image for the repository. Raise an exception if the repository
        isn't checked out."""
        return cast(Image, self.images.by_tag("HEAD", raise_on_none=True))

    @property
    def head(self) -> Optional[Image]:
        """Return the HEAD image for the repository or None if the repository isn't checked out."""
        return self.images.by_tag("HEAD", raise_on_none=False)

    @manage_audit
    def uncheckout(self, force: bool = False) -> None:
        """
        Deletes the schema that the repository is checked out into

        :param force: Discards all pending changes to the schema.
        """
        if not self.head:
            return
        if self.has_pending_changes():
            if not force:
                raise CheckoutError(
                    "{0} has pending changes! Pass force=True or do sgr checkout -f {0}:HEAD".format(
                        self.to_schema()
                    )
                )
            logging.warning("%s has pending changes, discarding...", self.to_schema())

        # Delete the schema and remove the HEAD tag
        self.delete(unregister=False, uncheckout=True)
        self.head.delete_tag("HEAD")

    def commit(
        self,
        image_hash: Optional[str] = None,
        comment: Optional[str] = None,
        snap_only: bool = False,
        chunk_size: Optional[int] = None,
        split_changeset: bool = False,
        extra_indexes: Optional[Dict[str, ExtraIndexInfo]] = None,
        in_fragment_order: Optional[Dict[str, List[str]]] = None,
        overwrite: bool = False,
    ) -> Image:
        """
        Commits all pending changes to a given repository, creating a new image.

        :param image_hash: Hash of the commit. Chosen by random if unspecified.
        :param comment: Optional comment to add to the commit.
        :param snap_only: If True, will store the table as a full snapshot instead of delta compression
        :param chunk_size: For tables that are stored as snapshots (new tables and where `snap_only` has been passed,
            the table will be split into fragments of this many rows.
        :param split_changeset: If True, splits the changeset into multiple fragments based on
            the PK regions spanned by the current table fragments. For example, if the original table
            consists of 2 fragments, first spanning rows 1-10000, second spanning rows 10001-20000 and the
            change alters rows 1, 10001 and inserts a row with PK 20001, this will record the change as
            3 fragments: one inheriting from the first original fragment, one inheriting from the second
            and a brand new fragment. This increases the number of fragments in total but means that fewer rows
            will need to be scanned to satisfy a query.
            If False, the changeset will be stored as a single fragment inheriting from the last fragment in the
            table.
        :param extra_indexes: Dictionary of {table: index_type: column: index_specific_kwargs}.
        :param in_fragment_order: Dictionary of {table: list of columns}. If specified, will
        sort the data inside each chunk by this/these key(s) for each table.
        :param overwrite: If an object already exists, will force recreate it.

        :return: The newly created Image object.
        """

        logging.info("Committing %s...", self.to_schema())

        self.object_engine.commit()

        # HEAD can be None (if this is the first commit in this repository)
        head = self.head
        if image_hash is None:
            # Generate a random hexadecimal hash for new images
            image_hash = "{:064x}".format(getrandbits(256))

        self.images.add(head.image_hash if head else None, image_hash, comment=comment)
        self._commit(
            head,
            image_hash,
            snap_only=snap_only,
            chunk_size=chunk_size,
            split_changeset=split_changeset,
            extra_indexes=extra_indexes,
            in_fragment_order=in_fragment_order,
            overwrite=overwrite,
        )

        set_head(self, image_hash)
        manage_audit_triggers(self.engine, self.object_engine)
        self.commit_engines()
        return self.images.by_hash(image_hash)

    def _commit(
        self,
        head: Optional[Image],
        image_hash: str,
        snap_only: bool = False,
        chunk_size: Optional[int] = None,
        split_changeset: bool = False,
        schema: str = None,
        extra_indexes: Optional[Dict[str, ExtraIndexInfo]] = None,
        in_fragment_order: Optional[Dict[str, List[str]]] = None,
        overwrite: bool = False,
    ) -> None:
        """
        Reads the recorded pending changes to all tables in a given checked-out image,
        conflates them and possibly stores them as new object(s) as follows:

            * If a table has been created or there has been a schema change, it's only stored as a full snapshot.
            * If a table hasn't changed since the last revision, no new objects are created and it's linked to the
                previous objects belonging to the last revision.
            * Otherwise, the table is stored as a conflated (1 change per PK) patch.
        """
        schema = schema or self.to_schema()
        extra_indexes: Dict[str, ExtraIndexInfo] = extra_indexes or {}
        in_fragment_order: Dict[str, List[str]] = in_fragment_order or {}
        chunk_size = chunk_size or int(get_singleton(CONFIG, "SG_COMMIT_CHUNK_SIZE"))

        changed_tables = self.object_engine.get_changed_tables(schema)
        tracked_tables = self.object_engine.get_tracked_tables()

        for table in self.object_engine.get_all_tables(schema):
            if self.object_engine.get_table_type(schema, table) == "VIEW":
                logging.warning(
                    "Table %s.%s is a view. Splitgraph currently doesn't "
                    "support views and this table will not be in the image.",
                    schema,
                    table,
                )
                continue

            try:
                table_info: Optional[Table] = head.get_table(table) if head else None
            except TableNotFoundError:
                table_info = None

            # Store as a full copy if this is a new table, there's been a schema change or we were told to.
            # Also store the full copy if the audit trigger on the table is missing: this indicates
            # that the table was dropped and recreated.
            new_schema = self.object_engine.get_full_table_schema(schema, table)
            if (
                not table_info
                or snap_only
                or not _schema_compatible(table_info.table_schema, new_schema)
                or (schema, table) not in tracked_tables
            ):
                self.objects.record_table_as_base(
                    self,
                    table,
                    image_hash,
                    chunk_size=chunk_size,
                    source_schema=schema,
                    extra_indexes=extra_indexes.get(table),
                    in_fragment_order=in_fragment_order.get(table),
                    overwrite=overwrite,
                )
                continue

            # If the table has changed, look at the audit log and store it as a delta.
            if table in changed_tables:
                self.objects.record_table_as_patch(
                    table_info,
                    schema,
                    image_hash,
                    new_schema_spec=new_schema,
                    split_changeset=split_changeset,
                    extra_indexes=extra_indexes.get(table),
                    in_fragment_order=in_fragment_order.get(table),
                    overwrite=overwrite,
                )
                continue

            # If the table wasn't changed, point the image to the old table
            self.objects.register_tables(
                self, [(image_hash, table, new_schema, table_info.objects)]
            )

        # Make sure that all pending changes have been discarded by this point (e.g. if we created just a snapshot for
        # some tables and didn't consume the audit log).
        # NB if we allow partial commits, this will have to be changed (only discard for committed tables).
        self.object_engine.discard_pending_changes(schema)

    def has_pending_changes(self) -> bool:
        """
        Detects if the repository has any pending changes (schema changes, table additions/deletions, content changes).
        """
        head = self.head
        if not head:
            # If the repo isn't checked out, no point checking for changes.
            return False
        for table in self.object_engine.get_all_tables(self.to_schema()):
            diff = self.diff(table, head.image_hash, None, aggregate=True)
            if diff != (0, 0, 0) and diff is not None:
                return True
        return False

    # --- TAG AND IMAGE MANAGEMENT ---

    def get_all_hashes_tags(self) -> List[Tuple[Optional[str], str]]:
        """
        Gets all tagged images and their hashes in a given repository.

        :return: List of (image_hash, tag)
        """
        return cast(
            List[Tuple[Optional[str], str]],
            self.engine.run_sql(
                select(
                    "get_tagged_images",
                    "image_hash, tag",
                    schema=SPLITGRAPH_API_SCHEMA,
                    table_args="(%s,%s)",
                ),
                (self.namespace, self.repository),
            ),
        )

    def set_tags(self, tags: Dict[str, Optional[str]]) -> None:
        """
        Sets tags for multiple images.

        :param tags: List of (image_hash, tag)
        """
        args = []
        for tag, image_id in tags.items():
            if tag != "HEAD":
                assert image_id is not None
                args.append((image_id, tag))
        set_tags_batch(self, args)

    def run_sql(
        self,
        sql: Union[Composed, str],
        arguments: Optional[Any] = None,
        return_shape: ResultShape = ResultShape.MANY_MANY,
    ) -> Any:
        """Execute an arbitrary SQL statement inside of this repository's checked out schema."""
        return self.object_engine.run_sql_in(self.to_schema(), sql, arguments, return_shape)

    def dump(self, stream: TextIOWrapper, exclude_object_contents: bool = False) -> None:
        """
        Creates an SQL dump with the metadata required for the repository and all of its objects.

        :param stream: Stream to dump the data into.
        :param exclude_object_contents: Only dump the metadata but not the actual object contents.
        """
        # First, go through the metadata tables required to reconstruct the repository.
        stream.write("--\n-- Images --\n--\n")
        self.engine.dump_table_sql(
            SPLITGRAPH_META_SCHEMA,
            "images",
            stream,
            where="namespace = %s AND repository = %s",
            where_args=(self.namespace, self.repository),
        )

        # Add objects (need to come before tables: we check that objects for inserted tables are registered.
        required_objects: Set[str] = set()
        for image in self.images:
            for table_name in image.get_tables():
                required_objects.update(image.get_table(table_name).objects)

        object_qual = (
            "object_id IN (" + ",".join(itertools.repeat("%s", len(required_objects))) + ")"
        )

        stream.write("\n--\n-- Objects --\n--\n")
        # To avoid conflicts, we just delete the object records if they already exist
        with self.engine.connection.cursor() as cur:
            for table_name in ("objects", "object_locations"):
                stream.write(
                    cur.mogrify(
                        SQL("DELETE FROM {}.{} WHERE ").format(
                            Identifier(SPLITGRAPH_META_SCHEMA), Identifier(table_name)
                        )
                        + SQL(object_qual),
                        list(required_objects),
                    ).decode("utf-8")
                )
                stream.write(";\n\n")
                self.engine.dump_table_sql(
                    SPLITGRAPH_META_SCHEMA,
                    table_name,
                    stream,
                    where=object_qual,
                    where_args=list(required_objects),
                )

        stream.write("\n--\n-- Tables --\n--\n")
        self.engine.dump_table_sql(
            SPLITGRAPH_META_SCHEMA,
            "tables",
            stream,
            where="namespace = %s AND repository = %s",
            where_args=(self.namespace, self.repository),
        )

        stream.write("\n--\n-- Tags --\n--\n")
        self.engine.dump_table_sql(
            SPLITGRAPH_META_SCHEMA,
            "tags",
            stream,
            where="namespace = %s AND repository = %s AND tag != 'HEAD'",
            where_args=(self.namespace, self.repository),
        )

        if not exclude_object_contents:
            with self.engine.connection.cursor() as cur:
                stream.write("\n--\n-- Object contents --\n--\n")

                # Finally, dump the actual objects
                for object_id in required_objects:
                    stream.write(
                        cur.mogrify(
                            SQL("DROP FOREIGN TABLE IF EXISTS {}.{};\n").format(
                                Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id)
                            )
                        ).decode("utf-8")
                    )
                    self.object_engine.dump_object(object_id, stream, schema=SPLITGRAPH_META_SCHEMA)
                    stream.write("\n")

    # --- IMPORTING TABLES ---

    @manage_audit
    def import_tables(
        self,
        tables: Sequence[str],
        source_repository: "Repository",
        source_tables: Sequence[str],
        image_hash: Optional[str] = None,
        foreign_tables: bool = False,
        do_checkout: bool = True,
        target_hash: Optional[str] = None,
        table_queries: Optional[Sequence[bool]] = None,
        parent_hash: Optional[str] = None,
        wrapper: Optional[str] = FDW_CLASS,
        skip_validation: bool = False,
    ) -> str:
        """
        Creates a new commit in target_repository with one or more tables linked to already-existing tables.
        After this operation, the HEAD of the target repository moves to the new commit and the new tables are
        materialized.

        :param tables: If not empty, must be the list of the same length as `source_tables` specifying names to store
            them under in the target repository.
        :param source_repository: Repository to import tables from.
        :param source_tables: List of tables to import. If empty, imports all tables.
        :param image_hash: Image hash in the source repository to import tables from.
            Uses the current source HEAD by default.
        :param foreign_tables: If True, copies all source tables to create a series of new snapshots instead of
            treating them as Splitgraph-versioned tables. This is useful for adding brand new tables
            (for example, from an FDW-mounted table).
        :param do_checkout: If False, doesn't check out the newly created image.
        :param target_hash: Hash of the new image that tables is recorded under. If None, gets chosen at random.
        :param table_queries: If not [], it's treated as a Boolean mask showing which entries in the `tables` list are
            instead SELECT SQL queries that form the target table. The queries have to be non-schema qualified and work
            only against tables in the source repository. Each target table created is the result of the respective SQL
            query. This is committed as a new snapshot.
        :param parent_hash: If not None, must be the hash of the image to base the new image on.
            Existing tables from the parent image are preserved in the new image. If None, the current repository
            HEAD is used.
        :param wrapper: Override the default class for the layered querying foreign data wrapper.
        :param skip_validation: Don't validate SQL used in import statements (used by the Splitfile executor that pre-formats the SQL).
        :return: Hash that the new image was stored under.
        """
        # Sanitize/validate the parameters and call the internal function.
        if table_queries is None:
            table_queries = []
        target_hash = target_hash or "{:064x}".format(getrandbits(256))

        image: Optional[Image]
        if not foreign_tables:
            image = (
                source_repository.images.by_hash(image_hash)
                if image_hash
                else source_repository.head_strict
            )
        else:
            image = None

        if not source_tables:
            if foreign_tables:
                source_tables = source_repository.object_engine.get_all_tables(
                    source_repository.to_schema()
                )
            else:
                assert image is not None
                source_tables = image.get_tables()
        if not tables:
            if table_queries:
                raise ValueError("target_tables has to be defined if table_queries is True!")
            tables = source_tables
        if not table_queries:
            table_queries = [False] * len(tables)
        if len(tables) != len(source_tables) or len(source_tables) != len(table_queries):
            raise ValueError("tables, source_tables and table_queries have mismatching lengths!")

        if parent_hash:
            existing_tables = self.images[parent_hash].get_tables()
        else:
            try:
                parent_hash = self.head.image_hash if self.head else None
            except RepositoryNotFoundError:
                parent_hash = None
            existing_tables = self.object_engine.get_all_tables(self.to_schema())
        clashing = [t for t in tables if t in existing_tables]
        if clashing:
            raise ValueError("Table(s) %r already exist(s) at %s!" % (clashing, self))

        return self._import_tables(
            image,
            tables,
            source_repository,
            target_hash,
            source_tables,
            do_checkout,
            table_queries,
            foreign_tables,
            parent_hash,
            wrapper,
            skip_validation,
        )

    def _import_tables(
        self,
        image: Optional[Image],
        tables: Sequence[str],
        source_repository: "Repository",
        target_hash: str,
        source_tables: Sequence[str],
        do_checkout: bool,
        table_queries: Sequence[bool],
        foreign_tables: bool,
        base_hash: Optional[str],
        wrapper: Optional[str],
        skip_validation: bool,
    ) -> str:
        # This importing route only supported between local repos.
        assert self.engine == source_repository.engine
        assert self.object_engine == source_repository.object_engine
        if do_checkout:
            self.object_engine.create_schema(self.to_schema())

        self.images.add(
            base_hash,
            target_hash,
            comment="Importing %s from %s" % (pluralise("table", len(tables)), source_repository),
        )

        # Materialize the actual tables in the target repository and register them.
        for source_table, target_table, is_query in zip(source_tables, tables, table_queries):
            # For foreign tables/SELECT queries, we define a new object/table instead.
            if is_query and not foreign_tables:
                # If we're importing a query from another Splitgraph image, we can use LQ to satisfy it.
                # This could get executed for the whole import batch as opposed to for every import query
                # but the overhead of setting up an LQ schema is fairly small.
                assert image is not None
                with image.query_schema(wrapper=wrapper) as tmp_schema:
                    self._import_new_table(
                        tmp_schema,
                        source_table,
                        target_hash,
                        target_table,
                        is_query,
                        do_checkout,
                        skip_validation,
                    )
            elif foreign_tables:
                self._import_new_table(
                    source_repository.to_schema(),
                    source_table,
                    target_hash,
                    target_table,
                    is_query,
                    do_checkout,
                )
            else:
                assert image is not None
                table_obj = image.get_table(source_table)
                self.objects.register_tables(
                    self, [(target_hash, target_table, table_obj.table_schema, table_obj.objects)]
                )
                if do_checkout:
                    table_obj.materialize(target_table, destination_schema=self.to_schema())
        # Register the existing tables at the new commit as well.
        if base_hash is not None:
            # Maybe push this into the tables API (currently have to make 2 queries)
            self.engine.run_sql(
                SQL(
                    """INSERT INTO {0}.tables (namespace, repository, image_hash,
                    table_name, table_schema, object_ids) (SELECT %s, %s, %s, table_name, table_schema, object_ids
                    FROM {0}.tables WHERE namespace = %s AND repository = %s AND image_hash = %s)"""
                ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
                (
                    self.namespace,
                    self.repository,
                    target_hash,
                    self.namespace,
                    self.repository,
                    base_hash,
                ),
            )
        if do_checkout:
            set_head(self, target_hash)
        return target_hash

    def _import_new_table(
        self,
        source_schema: str,
        source_table: str,
        target_hash: str,
        target_table: str,
        is_query: bool,
        do_checkout: bool,
        skip_validation: bool = False,
    ) -> List[str]:
        # First, import the query (or the foreign table) into a temporary table.
        tmp_object_id = get_temporary_table_id()
        if is_query:
            # is_query precedes foreign_tables: if we're importing using a query, we don't care if it's a
            # foreign table or not since we're storing it as a full snapshot.
            if not skip_validation:
                source_table = validate_import_sql(source_table)

            # Prepend query planner instructions to the query
            self.object_engine.run_sql_in(
                source_schema,
                SQL(get_singleton(CONFIG, "SG_LQ_TUNING"))
                + SQL("CREATE TABLE {}.{} AS ").format(
                    Identifier(SPLITGRAPH_META_SCHEMA), Identifier(tmp_object_id)
                )
                + SQL(source_table),
            )
        else:
            self.object_engine.copy_table(
                source_schema, source_table, SPLITGRAPH_META_SCHEMA, tmp_object_id
            )
        # This is kind of a waste: if the table is indeed new (and fits in one chunk),
        # the fragment manager will copy it over once again and give it the new object ID.
        # Maybe the fragment manager could rename the table in this case.
        try:
            # Commit so that the other threads can see the table.
            self.object_engine.commit()
            actual_objects = self.objects.record_table_as_base(
                self,
                target_table,
                target_hash,
                source_schema=SPLITGRAPH_META_SCHEMA,
                source_table=tmp_object_id,
            )
        finally:
            self.object_engine.delete_table(SPLITGRAPH_META_SCHEMA, tmp_object_id)
        if do_checkout:
            self.images.by_hash(target_hash).get_table(target_table).materialize(
                target_table, self.to_schema()
            )
        return actual_objects

    # --- SYNCING WITH OTHER REPOSITORIES ---

    def push(
        self,
        remote_repository: Optional["Repository"] = None,
        overwrite_objects: bool = False,
        reupload_objects: bool = False,
        overwrite_tags: bool = False,
        handler: str = "DB",
        handler_options: Optional[Dict[str, Any]] = None,
        single_image: Optional[str] = None,
    ) -> "Repository":
        """
        Inverse of ``pull``: Pushes all local changes to the remote and uploads new objects.

        :param remote_repository: Remote repository to push changes to. If not specified, the current
            upstream is used.
        :param handler: Name of the handler to use to upload objects. Use `DB` to push them to the remote or `S3`
            to store them in an S3 bucket.
        :param overwrite_objects: If True, will overwrite object metadata on the remote repository for existing objects.
        :param reupload_objects: If True, will reupload objects for which metadata is uploaded.
        :param overwrite_tags: If True, will overwrite existing tags on the remote repository.
        :param handler_options: Extra options to pass to the handler. For example, see
            :class:`splitgraph.hooks.s3.S3ExternalObjectHandler`.
        :param single_image: Limit the upload to a single image hash/tag.
        """
        remote_repository = remote_repository or self.upstream
        if not remote_repository:
            raise ValueError(
                "No remote repository specified and no upstream found for %s!" % self.to_schema()
            )

        try:
            _sync(
                target=remote_repository,
                source=self,
                download=False,
                overwrite_objects=overwrite_objects,
                reupload_objects=reupload_objects,
                overwrite_tags=overwrite_tags,
                handler=handler,
                handler_options=handler_options,
                single_image=single_image,
            )

            if not self.upstream:
                self.upstream = remote_repository
                logging.info("Setting upstream for %s to %s.", self, remote_repository)
        finally:
            # Don't commit the connection here: _sync is supposed to do it itself
            # after a successful push/pull.
            remote_repository.engine.close()
        return remote_repository

    def pull(
        self,
        download_all: Optional[bool] = False,
        overwrite_objects: bool = False,
        overwrite_tags: bool = False,
        single_image: Optional[str] = None,
    ) -> None:
        """
        Synchronizes the state of the local Splitgraph repository with its upstream, optionally downloading all new
        objects created on the remote.

        :param download_all: If True, downloads all objects and stores them locally. Otherwise, will only download
            required objects when a table is checked out.
        :param overwrite_objects: If True, will overwrite object metadata on the local repository for existing objects.
        :param overwrite_tags: If True, will overwrite existing tags.
        :param single_image: Limit the download to a single image hash/tag.
        """
        if not self.upstream:
            raise ValueError("No upstream found for repository %s!" % self.to_schema())

        clone(
            remote_repository=self.upstream,
            local_repository=self,
            download_all=download_all,
            overwrite_objects=overwrite_objects,
            overwrite_tags=overwrite_tags,
            single_image=single_image,
        )

    def diff(
        self,
        table_name: str,
        image_1: Union[Image, str],
        image_2: Optional[Union[Image, str]],
        aggregate: bool = False,
    ) -> Union[bool, Tuple[int, int, int], List[Tuple[bool, Tuple]], None]:
        """
        Compares the state of a table in different images by materializing both tables into a temporary space
        and comparing them row-to-row.

        :param table_name: Name of the table.
        :param image_1: First image hash / object. If None, uses the state of the current staging area.
        :param image_2: Second image hash / object. If None, uses the state of the current staging area.
        :param aggregate: If True, returns a tuple of integers denoting added, removed and updated rows between
            the two images.
        :return: If the table doesn't exist in one of the images, returns True if it was added and False if it was
            removed. If `aggregate` is True, returns the aggregation of changes as specified before.
            Otherwise, returns a list of changes where each change is a tuple of
            `(True for added, False for removed, row contents)`.
        """

        if isinstance(image_1, str):
            image_1 = self.images.by_hash(image_1)
        if isinstance(image_2, str):
            image_2 = self.images.by_hash(image_2)

        # If the table is of an unsupported type (e.g. a view), ignore it
        if (
            self.object_engine.table_exists(self.to_schema(), table_name)
            and self.object_engine.get_table_type(self.to_schema(), table_name) == "VIEW"
        ):
            return None

        # If the table doesn't exist in the first or the second image, short-circuit and
        # return the bool.
        if not table_exists_at(self, table_name, image_1):
            return True
        if not table_exists_at(self, table_name, image_2):
            return False

        # Special case: if diffing HEAD and staging (with aggregation), we can return that directly.
        if (
            image_1 == self.head
            and image_2 is None
            and aggregate
            and (self.to_schema(), table_name) in self.object_engine.get_tracked_tables()
        ):
            return aggregate_changes(
                cast(
                    List[Tuple[int, int]],
                    self.object_engine.get_pending_changes(
                        self.to_schema(), table_name, aggregate=True
                    ),
                )
            )

        # If the table is the same in the two images, short circuit as well.
        if image_2 is not None and set(image_1.get_table(table_name).objects) == set(
            image_2.get_table(table_name).objects
        ):
            return [] if not aggregate else (0, 0, 0)

        # Materialize both tables and compare them side-by-side.
        # TODO we can aggregate chunks in a similar way that LQ does it.
        return slow_diff(self, table_name, _hash(image_1), _hash(image_2), aggregate)


def import_table_from_remote(
    remote_repository: "Repository",
    remote_tables: List[str],
    remote_image_hash: str,
    target_repository: "Repository",
    target_tables: List[Any],
    target_hash: str = None,
) -> None:
    """
    Shorthand for importing one or more tables from a yet-uncloned remote. Here, the remote image hash is required,
    as otherwise we aren't necessarily able to determine what the remote head is.

    :param remote_repository: Remote Repository object
    :param remote_tables: List of remote tables to import
    :param remote_image_hash: Image hash to import the tables from
    :param target_repository: Target repository to import the tables to
    :param target_tables: Target table aliases
    :param target_hash: Hash of the image that's created with the import. Default random.
    """

    # In the future, we could do some vaguely intelligent interrogation of the remote to directly copy the required
    # metadata (object locations and relationships) into the local mountpoint. However, since the metadata is fairly
    # lightweight (we never download unneeded objects), we just clone it into a temporary mountpoint,
    # do the import into the target and destroy the temporary mp.
    tmp_mountpoint = Repository(
        namespace=remote_repository.namespace,
        repository=remote_repository.repository + "_clone_tmp",
    )

    clone(remote_repository, local_repository=tmp_mountpoint, download_all=False)
    target_repository.import_tables(
        target_tables,
        tmp_mountpoint,
        remote_tables,
        image_hash=remote_image_hash,
        target_hash=target_hash,
    )

    tmp_mountpoint.delete()
    target_repository.commit_engines()


def table_exists_at(
    repository: "Repository", table_name: str, image: Optional[Image] = None
) -> bool:
    """Determines whether a given table exists in a Splitgraph image without checking it out. If `image_hash` is None,
    determines whether the table exists in the current staging area."""
    if image is None:
        return repository.object_engine.table_exists(repository.to_schema(), table_name)
    else:
        try:
            image.get_table(table_name)
            return True
        except TableNotFoundError:
            return False


def _sync(
    target: "Repository",
    source: "Repository",
    download: bool = True,
    overwrite_objects: bool = False,
    reupload_objects: bool = False,
    overwrite_tags: bool = False,
    handler: str = "DB",
    handler_options: Optional[Dict[str, Any]] = None,
    single_image: Optional[str] = None,
) -> None:
    """
    Generic routine for syncing two repositories: fetches images, hashes, objects and tags
    on `source` that don't exist in `target`.

    Common code between push and pull, since the only difference between those routines is that
    uploading and downloading objects are different operations.

    :param target: Target Repository object
    :param source: Source Repository object
    :param download: If True, uses the download routines to download physical objects to self.
        If False, uses the upload routines to get `source` to upload physical objects to self / external.
    :param overwrite_objects: If True, overwrites remote object metadata for the target repository.
    :param reupload_objects: If True, will reupload objects for which metadata is uploaded.
    :param overwrite_tags: If True, overwrites tags for all (if syncing the whole repository)
        or the single image.
    :param handler: Upload handler
    :param handler_options: Upload handler options
    :param single_image: Limit the download/upload to a single image hash/tag.
    """
    if handler_options is None:
        handler_options = {}

    # Get the remote log and the list of objects we need to fetch.
    logging.info("Gathering remote metadata...")

    partial_upload_failure: Optional[IncompleteObjectUploadError] = None

    # Run a read-only transaction that gets images/objects we need to download
    new_images, table_meta, object_locations, object_meta, tags = gather_sync_metadata(
        target,
        source,
        overwrite_objects=overwrite_objects,
        single_image=single_image,
        overwrite_tags=overwrite_tags,
    )
    if not new_images and not object_meta and not object_locations and not tags:
        logging.info("No image/object metadata to pull.")
        return

    try:
        if download:
            target.objects.register_objects(list(object_meta.values()))
            target.objects.register_object_locations(object_locations)

            # Don't check anything out, keep the repo bare.
            set_head(target, None)
        else:
            # If we're overwriting, upload all objects that we're pushing metadata for
            new_objects = target.objects.get_new_objects(list(object_meta.keys()))

            objects_to_push = (
                new_objects if not reupload_objects else [o for o, om in object_meta.items()]
            )

            # Transaction handling here is finicky as we don't want to hold an open transaction
            # to the registry, blocking other clients. gather_sync_metadata was a self-contained
            # read-only transaction. The S3 uploader runs a self-contained RO transaction as well
            # to find out the URLs it's uploading objects to.

            try:
                new_uploads = source.objects.upload_objects(
                    target.objects,
                    objects_to_push,
                    handler=handler,
                    handler_params=handler_options,
                )
                new_locations = [(o, u, handler) for o, u in new_uploads if u]
                successful = [o for o, _ in new_uploads]
            except IncompleteObjectUploadError as e:
                partial_upload_failure = e
                new_locations = [
                    (o, u, handler)
                    for o, u in zip(e.successful_objects, e.successful_object_urls)
                    if u
                ]
                successful = e.successful_objects

            # Here we have to register the new objects after the upload but before we store
            # their external location (as access control for object_locations relies on the
            # object metadata being in place).
            # In addition, register all objects we had in object_meta apart from those that
            # we tried to upload and failed (if we're pushing overwriting object metadata,
            # object_meta might have more objects than what we actually intended to upload).

            # Note that this implicitly opens a short-lived write transaction to the target
            # (we register objects, their locations and then the actual tables and images
            # when we leave this if).
            target.objects.register_objects(
                [v for k, v in object_meta.items() if k not in objects_to_push or k in successful],
                namespace=target.namespace,
            )

            target.objects.register_object_locations(list(set(object_locations + new_locations)))
            source.objects.register_object_locations(new_locations)

            if partial_upload_failure:
                # If the upload has failed or was cancelled for some reason, still register
                # successful objects (give the caller a chance to continue the upload) but
                # not images or tables -- bail out now.
                target.commit_engines()
                source.commit_engines()
                if partial_upload_failure.reason:
                    raise partial_upload_failure.reason
                else:
                    raise partial_upload_failure

        # Register the new images / tables / tags.
        target.images.add_batch(new_images)
        target.objects.register_tables(target, table_meta)
        target.set_tags(tags)
    except Exception:
        logging.exception("Error during repository sync")
        target.rollback_engines()
        source.rollback_engines()
        raise

    target.commit_engines()
    source.commit_engines()

    logging.info(
        "%s metadata for %s, %s, %s and %s.",
        ("Fetched" if download else "Uploaded"),
        pluralise("image", len(new_images)),
        pluralise("table", len(table_meta)),
        pluralise("object", len(object_meta)),
        pluralise("tag", len([t for t in tags if t != "HEAD"])),
    )


def clone(
    remote_repository: Union["Repository", str],
    local_repository: Optional["Repository"] = None,
    overwrite_objects: bool = False,
    overwrite_tags: bool = False,
    download_all: Optional[bool] = False,
    single_image: Optional[str] = None,
) -> "Repository":
    """
    Clones a remote Splitgraph repository or synchronizes remote changes with the local ones.

    If the target repository has no set upstream engine, the source repository becomes its upstream.

    :param remote_repository: Remote Repository object to clone or the repository's name. If a name is passed,
        the repository will be looked up on the current lookup path in order to find the engine the repository
        belongs to.
    :param local_repository: Local repository to clone into. If None, uses the same name as the remote.
    :param download_all: If True, downloads all objects and stores them locally. Otherwise, will only download required
        objects when a table is checked out.
    :param overwrite_objects: If True, will overwrite object metadata on the local repository for existing objects.
    :param overwrite_tags: If True, will overwrite existing tags.
    :param single_image: If set, only get a single image with this hash/tag from the source.
    :return: A locally cloned Repository object.
    """
    if isinstance(remote_repository, str):
        remote_repository = lookup_repository(remote_repository, include_local=False)

    # Repository engine should be local by default
    if not local_repository:
        local_repository = Repository(remote_repository.namespace, remote_repository.repository)

    _sync(
        local_repository,
        remote_repository,
        download=True,
        overwrite_objects=overwrite_objects,
        overwrite_tags=overwrite_tags,
        single_image=single_image,
    )

    # Perform the optional download of all objects as a final step (normally we do it when the user
    # tries to check out a revision) and do it using the object manager as it has some handling
    # of error cases.
    if download_all:
        local_om = local_repository.objects
        with local_om.ensure_objects(
            table=None,
            objects=local_om.get_objects_for_repository(
                local_repository,
                image_hash=local_repository.images[single_image].image_hash
                if single_image
                else None,
            ),
            upstream_manager=remote_repository.objects,
        ):
            pass

    if not local_repository.upstream:
        local_repository.upstream = remote_repository

    return local_repository


def _hash(image: Optional[Image]) -> Optional[str]:
    return image.image_hash if image is not None else None


def _schema_compatible(old: TableSchema, new: TableSchema) -> bool:
    """
    Determines whether a table with a schema change can be delta compressed
    using data in the audit triggers.

    :param old: Old table schema
    :param new: New table schema
    :return: Boolean
    """

    # Currently we don't support schema changes at all but we do delta compress if
    # just the comments on some of the table's columns have changed.
    return [t[:4] for t in old] == [t[:4] for t in new]



import logging
import os
import platform
from pathlib import Path, PureWindowsPath
from typing import Any, Dict, Optional, cast
from urllib.parse import urlparse

import click
from tqdm import tqdm

from splitgraph.__version__ import __version__
from splitgraph.config import CONFIG, SG_CMD_ASCII, get_singleton
from splitgraph.config.management import patch_and_save_config
from splitgraph.exceptions import DockerUnavailableError, EngineSetupError
from splitgraph.utils.docker import copy_to_container, get_docker_client

DEFAULT_ENGINE = "default"


def inject_config_into_engines(engine_prefix, config_path):
    """
    Copy the current config into all engines that are managed by `sgr engine`. This
    is so that the engine has the right credentials and settings for when we do
    layered querying (a Postgres client queries the engine directly and it
    has to download objects etc).

    :param engine_prefix: Prefix for Docker containers that are considered to be engines
    :param config_path: Path to the config file.
    """
    engine_containers = list_engines(engine_prefix, include_all=True)
    if engine_containers:
        logging.info("Copying the config file at %s into all current engines", config_path)
        for container in engine_containers:
            copy_to_container(container, config_path, "/.sgconfig")
            logging.info("Config updated for container %s", container.name)


def _get_container_name(engine_name: str) -> str:
    from splitgraph.config import CONFIG

    return "%s%s" % (CONFIG["SG_ENGINE_PREFIX"], engine_name)


def _get_data_volume_name(engine_name: str) -> str:
    from splitgraph.config import CONFIG

    return "%s%s_data" % (CONFIG["SG_ENGINE_PREFIX"], engine_name)


def _get_metadata_volume_name(engine_name: str) -> str:
    from splitgraph.config import CONFIG

    return "%s%s_metadata" % (CONFIG["SG_ENGINE_PREFIX"], engine_name)


def _convert_source_path(path: str) -> str:
    """If we're passed a Windows-style path to mount into the container,
    we need to convert it to Unix-style for when the user is running Docker
    in a docker-machine VM on Windows. Docker mounts C:\\Users into /c/Users on the
    machine, so anything that's a subdirectory of that is available for bind mounts."""
    pathobj = Path(path)
    if isinstance(pathobj, PureWindowsPath):
        path = str(pathobj.as_posix())
    if path[1] == ":":
        # If the path has a drive letter (C:/Users/... is recognized as a PosixPath on Linux
        # but that's not what we want), we need to convert it into the /c/Users/... mount
        if not path.lower().startswith("c:/users"):
            logging.warning(
                "Windows-style path %s might not be available for bind mounting in Docker", path
            )
        drive_letter = path[0].lower()
        return "/" + drive_letter + path[2:]
    return path


@click.group(name="engine")
def engine_c():
    """Manage running Splitgraph engines. This is a wrapper around the relevant Docker commands."""


def list_engines(prefix, include_all=False, unavailable_ok=True):
    try:
        client = get_docker_client()
    except DockerUnavailableError:
        if not unavailable_ok:
            raise
        logging.warning(
            "Could not connect to the Docker daemon to enumerate engines managed by sgr. This is fine if you're managing the engine yourself with Docker or Compose."
        )
        return []
    containers = client.containers.list(all=include_all)

    return [c for c in containers if c.name.startswith(prefix)]


@click.command(name="list")
@click.option(
    "-a", "--include-all", is_flag=True, default=False, help="Include stopped engine containers."
)
def list_engines_c(include_all):
    """List Splitgraph engines.

    This only lists Docker containers that were created by sgr engine
    (whose names start with `splitgraph_engine_`. To manage other engines,
    use Docker CLI directly.
    """
    from tabulate import tabulate

    from splitgraph.config import CONFIG

    containers = list_engines(include_all=include_all, prefix=CONFIG["SG_ENGINE_PREFIX"])
    if containers:
        our_containers = []
        for container in containers:
            engine_name = container.name[len(CONFIG["SG_ENGINE_PREFIX"]) :]
            ports = container.attrs["NetworkSettings"]["Ports"]
            ports = ",".join("%s -> %s" % i for i in ports.items())
            our_containers.append((engine_name, container.short_id, container.status, ports))

        click.echo(
            tabulate(
                our_containers, headers=("Name", "Docker ID", "Status", "ports"), tablefmt="plain"
            )
        )


def _pretty_pull(client, image):
    # Use the details from the low-level docker API to give us a pull progressbar

    use_ascii = True if platform.system() == "Windows" else SG_CMD_ASCII

    with tqdm(
        total=1,
        desc="Downloading",
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
        position=0,
        ascii=use_ascii,
    ) as download_bar:
        with tqdm(
            total=1,
            desc="Extracting",
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            position=1,
            ascii=use_ascii,
        ) as extract_bar:
            download_progress: Dict[str, Dict[str, Any]] = {}
            extract_progress: Dict[str, Dict[str, Any]] = {}
            for progress in client.api.pull(image, stream=True, decode=True):
                progress_text = "%s: %s" % (progress.get("id", ""), progress.get("status", ""))
                if progress["status"] == "Downloading":
                    _update_bar(progress, progress_text, download_bar, download_progress)
                elif progress["status"] == "Extracting":
                    _update_bar(progress, progress_text, extract_bar, extract_progress)


def _update_bar(
    progress: Dict[str, Any],
    progress_text: str,
    bar: tqdm,
    progress_data: Dict[str, Dict[str, Any]],
):
    if progress.get("progressDetail"):
        progress_data[progress["id"]] = progress["progressDetail"]

        bar.total = sum(p["total"] for p in progress_data.values())
        bar.update(sum(p["current"] for p in progress_data.values()) - bar.n)
    bar.set_description_str(progress_text)


@click.command(name="add")
@click.option(
    "-i",
    "--image",
    default="splitgraph/engine:%s" % __version__,
    help="Docker image with the Splitgraph engine",
)
@click.option("-p", "--port", type=int, default=6432, help="Port to start the engine on")
@click.option("-u", "--username", default="sgr")
@click.option("--no-init", default=False, help="Don't run `sgr init` on the engine", is_flag=True)
@click.option(
    "--no-sgconfig", default=False, help="Don't add the engine to .sgconfig", is_flag=True
)
@click.option(
    "--inject-source",
    default=False,
    help="Inject the current Splitgraph source code into the engine using Docker bind mounts",
    is_flag=True,
)
@click.option("--no-pull", default=False, help="Don't pull the Docker image", is_flag=True)
@click.option(
    "--set-default",
    default=False,
    help="Set the engine as the default engine in the config regardless of its name",
    is_flag=True,
)
@click.option(
    "--cap-add", help="Add kernel capabilities to the engine container", multiple=True, default=[]
)
@click.argument("name", default=DEFAULT_ENGINE)
@click.password_option()
def add_engine_c(
    image,
    port,
    username,
    no_init,
    no_sgconfig,
    inject_source,
    no_pull,
    name,
    password,
    set_default,
    cap_add,
):
    """
    Create and start a Splitgraph engine.

    This will pull the Splitgraph engine image, start it, create a Postgres user and initialize
    the engine.

    This also creates Docker volumes required to persist data/metadata.

    The engine Docker container by default will be named `splitgraph_engine_default` and
    its data and metadata volumes will have names `splitgraph_engine_default_data` and
    `splitgraph_engine_default_metadata`.
    """
    import docker
    from docker.types import Mount

    from splitgraph.config import CONFIG
    from splitgraph.engine.postgres.engine import PostgresEngine

    client = get_docker_client()

    if not no_pull:
        click.echo("Pulling image %s..." % image)
        _pretty_pull(client, image)

    container_name = _get_container_name(name)
    data_name = _get_data_volume_name(name)
    metadata_name = _get_metadata_volume_name(name)

    # Setup required mounts for data/metadata
    data_volume = Mount(target="/var/lib/splitgraph/objects", source=data_name, type="volume")
    metadata_volume = Mount(target="/var/lib/postgresql/data", source=metadata_name, type="volume")
    mounts = [data_volume, metadata_volume]

    click.echo("Creating container %s." % container_name)
    click.echo("Data volume: %s." % data_name)
    click.echo("Metadata volume: %s." % metadata_name)

    if inject_source:
        source_path = _convert_source_path(
            os.getenv(
                "SG_SOURCE_ROOT", os.path.abspath(os.path.join(os.path.dirname(__file__), "../"))
            )
        )
        source_volume = Mount(target="/splitgraph/splitgraph", source=source_path, type="bind")
        mounts.append(source_volume)
        click.echo("Source path: %s" % source_path)

    try:
        container = client.containers.run(
            image=image,
            detach=True,
            name=container_name,
            ports={"5432/tcp": port},
            mounts=mounts,
            environment={
                "POSTGRES_USER": username,
                "POSTGRES_PASSWORD": password,
                "POSTGRES_DB": "splitgraph",
                # Actual config to be injected later
                "SG_CONFIG_FILE": "/.sgconfig",
            },
            cap_add=cap_add,
        )
    except docker.errors.APIError as e:
        if "port is already allocated" in str(e):
            # Delete the container so that this operation is idempotent.
            container = client.containers.get(container_name)
            container.remove(force=True)

            raise EngineSetupError(
                f"Port {port} is already allocated. "
                "Note that Splitgraph does not currently support installation on "
                "existing PostgreSQL instances."
            ) from e
        if "Conflict. The container name" in str(e) and "is already in use by container" in str(e):
            raise EngineSetupError(
                f"Docker container with Splitgraph engine {container_name} already exists. Delete it with docker rm -f {container_name}.",
            ) from e
        raise

    click.echo("Container created, ID %s" % container.short_id)

    # Extract the host that we can reach the container on
    # (might be different from localhost if docker-machine is used)
    hostname = urlparse(client.api.base_url).hostname

    conn_params: Dict[str, str] = {
        "SG_ENGINE_HOST": hostname,
        "SG_ENGINE_PORT": str(port),
        # Even if the engine is exposed on a different port on the host,
        # need to make sure that it uses the default 5432 port to connect
        # to itself.
        "SG_ENGINE_FDW_HOST": "localhost",
        "SG_ENGINE_FDW_PORT": "5432",
        "SG_ENGINE_USER": username,
        "SG_ENGINE_PWD": password,
        "SG_ENGINE_DB_NAME": "splitgraph",
        "SG_ENGINE_POSTGRES_DB_NAME": "postgres",
        "SG_ENGINE_ADMIN_USER": username,
        "SG_ENGINE_ADMIN_PWD": password,
    }

    config_patch: Dict[str, Any]
    if not no_sgconfig:
        if name != DEFAULT_ENGINE and not set_default:
            config_patch = {"remotes": {name: conn_params}}
        else:
            config_patch = conn_params

        config_path = patch_and_save_config(CONFIG, config_patch)
    else:
        config_path = CONFIG["SG_CONFIG_FILE"]

    if not no_init:
        engine = PostgresEngine(
            name=name, conn_params=cast(Optional[Dict[str, Optional[str]]], conn_params)
        )
        engine.initialize()
        engine.commit()
        click.echo("Engine initialized successfully.")

    inject_config_into_engines(CONFIG["SG_ENGINE_PREFIX"], config_path)
    click.echo("Done.")


@click.command(name="stop")
@click.argument("name", default=DEFAULT_ENGINE)
def stop_engine_c(name):
    """Stop a Splitgraph engine.

    This is a wrapper around the corresponding Docker command.
    """

    client = get_docker_client()
    container_name = _get_container_name(name)
    container = client.containers.get(container_name)

    click.echo("Stopping Splitgraph engine %s..." % name)
    container.stop()
    click.echo("Engine stopped.")


@click.command(name="start")
@click.argument("name", default=DEFAULT_ENGINE)
def start_engine_c(name):
    """Start a Splitgraph engine.

    This is a wrapper around the corresponding Docker command.
    """

    client = get_docker_client()
    container_name = _get_container_name(name)
    container = client.containers.get(container_name)

    click.echo("Starting Splitgraph engine %s..." % name)
    container.start()
    click.echo("Engine started.")


@click.command(name="delete")
@click.option("-y", "--yes", default=False, is_flag=True, help="Do not prompt for confirmation.")
@click.option(
    "-f", "--force", default=False, is_flag=True, help="Delete the engine anyway if it's running."
)
@click.option(
    "-v",
    "--with-volumes",
    default=False,
    is_flag=True,
    help="Include the engine's volumes (if not specified, volumes will be reattached when an engine"
    " with the same name is created).",
)
@click.argument("name", default=DEFAULT_ENGINE)
def delete_engine_c(yes, force, with_volumes, name):
    """Delete the Splitgraph engine container."""

    client = get_docker_client()
    container_name = _get_container_name(name)
    container = client.containers.get(container_name)

    click.echo(
        "Splitgraph engine %s (container ID %s)%s will be deleted."
        % (name, container.short_id, (", together with all data," if with_volumes else ""))
    )
    if not yes:
        click.confirm("Continue? ", abort=True)

    container.remove(force=force)
    click.echo("Splitgraph engine %s has been removed." % name)

    if with_volumes:
        metadata_volume = client.volumes.get(_get_metadata_volume_name(name))
        data_volume = client.volumes.get(_get_data_volume_name(name))

        metadata_volume.remove()
        data_volume.remove()
        click.echo(
            "Volumes %s and %s have been removed." % (metadata_volume.name, data_volume.name)
        )


@click.command("log")
@click.option("-f", "--follow", is_flag=True, help="Stream logs")
@click.argument("name", default=DEFAULT_ENGINE)
def log_engine_c(name, follow):
    """Get logs from a Splitgraph engine."""

    client = get_docker_client()
    container_name = _get_container_name(name)
    container = client.containers.get(container_name)

    if follow:
        for line in container.logs(stream=True):
            click.echo(line, nl=False)
    else:
        click.echo(container.logs())


@click.command("configure")
@click.argument("name", default=DEFAULT_ENGINE)
def configure_engine_c(name):
    """Inject a configuration file into an engine.

    This copies the current .sgconfig file (pointed to by SG_CONFIG_FILE) into
    the engine container, making it use that configuration for
    when it's queried through an application other than the sgr client
    (layered querying)."""

    client = get_docker_client()
    container_name = _get_container_name(name)
    container = client.containers.get(container_name)

    copy_to_container(container, get_singleton(CONFIG, "SG_CONFIG_FILE"), "/.sgconfig")
    logging.info("Config updated for container %s", container.name)


@click.command("version")
@click.argument("name", default=DEFAULT_ENGINE)
def version_engine_c(name):
    """Get version of Splitgraph engine."""
    from splitgraph.engine import get_engine

    if name == DEFAULT_ENGINE:
        engine = get_engine()
    else:
        engine = get_engine(name)

    version = engine.splitgraph_version
    if version:
        click.echo("Splitgraph Engine %s" % version)


@click.command("upgrade")
@click.option(
    "-i",
    "--image",
    default="splitgraph/engine:%s" % __version__,
    help="Docker image with the Splitgraph engine",
)
@click.option("--no-pull", is_flag=True, help="Don't pull the new engine image")
@click.argument("name", default=DEFAULT_ENGINE)
@click.pass_context
def upgrade_engine_c(ctx, image, no_pull, name):
    """Upgrade a Splitgraph engine.

    This consists of shutting down the current Splitgraph engine,
    deleting its Docker container (keeping the actual data and
    metadata volumes intact), creating a container based on a newer
    image and finally reinitializing the engine to perform needed migrations.
    """
    from splitgraph.engine import get_engine

    # Get reference to engine to extract its connection params
    if name == DEFAULT_ENGINE:
        engine = get_engine()
    else:
        engine = get_engine(name)

    username = engine.conn_params["SG_ENGINE_USER"]
    password = engine.conn_params["SG_ENGINE_PWD"]
    port = engine.conn_params["SG_ENGINE_PORT"]

    # Stop the engine
    ctx.invoke(stop_engine_c, name=name)

    # Delete the container
    ctx.invoke(delete_engine_c, name=name, yes=True)

    # Create and start new engine
    ctx.invoke(
        add_engine_c,
        image=image,
        port=port,
        username=username,
        password=password,
        no_sgconfig=True,
        no_pull=no_pull,
        name=name,
    )

    version = engine.splitgraph_version
    if version:
        click.echo("Upgraded engine %s to %s" % (name, version))


engine_c.add_command(list_engines_c)
engine_c.add_command(add_engine_c)
engine_c.add_command(stop_engine_c)
engine_c.add_command(start_engine_c)
engine_c.add_command(delete_engine_c)
engine_c.add_command(upgrade_engine_c)
engine_c.add_command(log_engine_c)
engine_c.add_command(configure_engine_c)
engine_c.add_command(version_engine_c)



from abc import abstractmethod
from datetime import datetime as dt
from typing import Dict, List, Optional, Tuple, Union

from psycopg2.sql import SQL, Identifier

from splitgraph.core.image import Image
from splitgraph.core.repository import Repository
from splitgraph.core.sql import POSTGRES_MAX_IDENTIFIER
from splitgraph.core.types import TableColumn, TableSchema
from splitgraph.engine.postgres.engine import PsycopgEngine
from splitgraph.exceptions import CheckoutError


def schema_compatible(source_schema: TableSchema, target_schema: TableSchema) -> bool:
    """Quick check to see if a dataframe with target_schema can be written into source_schema.
    There are some implicit type conversions that SQLAlchemy/Pandas can do so we don't want to immediately fail
    if the column types aren't exactly the same (eg bigint vs numeric etc). Most errors should be caught by PG itself.

    Schema is a list of (ordinal, name, type, is_pk).
    """
    if len(source_schema) != len(target_schema):
        return False

    # Only check column names
    return all(
        col1.name == col2.name for col1, col2 in zip(sorted(source_schema), sorted(target_schema))
    )


def merge_tables(
    engine: "PsycopgEngine",
    source_schema: str,
    source_table: str,
    source_schema_spec: TableSchema,
    target_schema: str,
    target_table: str,
    target_schema_spec: TableSchema,
):

    # Construct an upsert query: INSERT INTO target_table (cols) (SELECT cols FROM tmp_table) ON CONFLICT (pk_cols)
    #                            DO UPDATE SET (non_pk_cols) = (EXCLUDED.non_pk_cols)
    pk_cols = [c[1] for c in target_schema_spec if c[3]]
    non_pk_cols = [c[1] for c in target_schema_spec if not c[3]]
    # Cast cols for when we're inserting things that Pandas detected as strings into columns with
    # datetimes/ints/etc
    cols_sql = SQL(",").join(Identifier(c[1]) for c in target_schema_spec)
    cols_sql_cast = SQL(",").join(
        Identifier(s[1]) + SQL("::" + t[2]) for s, t in zip(source_schema_spec, target_schema_spec)
    )
    query = (
        SQL("INSERT INTO {}.{} (").format(Identifier(target_schema), Identifier(target_table))
        + cols_sql
        + SQL(") (SELECT ")
        + cols_sql_cast
        + SQL(" FROM {}.{})").format(Identifier(source_schema), Identifier(source_table))
        + SQL(" ON CONFLICT (")
        + SQL(",").join(map(Identifier, pk_cols))
        + SQL(")")
    )
    if non_pk_cols:
        if len(non_pk_cols) > 1:
            query += (
                SQL(" DO UPDATE SET (")
                + SQL(",").join(map(Identifier, non_pk_cols))
                + SQL(") = (")
                + SQL(",").join([SQL("EXCLUDED.") + Identifier(c) for c in non_pk_cols])
                + SQL(")")
            )
        else:
            # otherwise, we get a "source for a multiple-column UPDATE item must be a sub-SELECT or ROW()
            # expression" error from pg.
            query += SQL(" DO UPDATE SET {0} = EXCLUDED.{0}").format(Identifier(non_pk_cols[0]))
    else:
        query += SQL(" DO NOTHING")
    engine.run_sql(query)


class IngestionAdapter:
    @abstractmethod
    def create_ingestion_table(self, data, engine, schema: str, table: str, **kwargs):
        pass

    @abstractmethod
    def data_to_new_table(
        self, data, engine, schema: str, table: str, no_header: bool = True, **kwargs
    ):
        pass

    @abstractmethod
    def query_to_data(self, engine, query: str, schema: Optional[str] = None, **kwargs):
        pass

    def to_table(
        self,
        data,
        repository: "Repository",
        table: str,
        if_exists: str = "patch",
        schema_check: bool = True,
        no_header: bool = False,
        **kwargs,
    ):
        tmp_schema = repository.to_schema()

        if not repository.head:
            raise CheckoutError("Repository %s isn't checked out!" % tmp_schema)

        table_exists = repository.object_engine.table_exists(tmp_schema, table)
        if not table_exists or (table_exists and if_exists == "replace"):
            self.create_ingestion_table(data, repository.object_engine, tmp_schema, table, **kwargs)

            if if_exists == "replace":
                # Currently, if a table is dropped and recreated with the same schema,
                # Splitgraph has no way of finding that out. So what we do is drop the last column
                # in the schema and then create it again with the same type: this increments the
                # ordinal on the column, which Splitgraph detects as a schema change.
                #
                # This is a (very ugly) hack but the root cause fix is not straightforward.

                table_schema = repository.object_engine.get_full_table_schema(tmp_schema, table)
                repository.object_engine.run_sql_in(
                    tmp_schema,
                    SQL(
                        "ALTER TABLE {0} DROP COLUMN {1};ALTER TABLE {0} ADD COLUMN {1} %s"
                        % table_schema[-1].pg_type
                    ).format(Identifier(table), Identifier(table_schema[-1].name)),
                )

            self.data_to_new_table(
                data, repository.object_engine, tmp_schema, table, no_header, **kwargs
            )
            repository.commit_engines()
            return

        # If we've reached this point, the table exists and we're patching values into it.
        # Ingest the table into a temporary location.
        tmp_table = "sg_tmp_ingestion" + table
        self.create_ingestion_table(data, repository.object_engine, tmp_schema, tmp_table, **kwargs)

        try:
            source_schema = repository.engine.get_full_table_schema(tmp_schema, tmp_table)
            target_schema = repository.engine.get_full_table_schema(tmp_schema, table)
            if schema_check and not schema_compatible(source_schema, target_schema):
                raise ValueError(
                    "Schema changes are unsupported with if_exists='patch'!"
                    "\nSource schema: %r\nTarget schema: %r" % (source_schema, target_schema)
                )

            self.data_to_new_table(
                data, repository.engine, tmp_schema, tmp_table, no_header=no_header
            )

            merge_tables(
                repository.engine,
                source_schema=tmp_schema,
                source_table=tmp_table,
                source_schema_spec=source_schema,
                target_schema=tmp_schema,
                target_table=table,
                target_schema_spec=target_schema,
            )
            repository.commit_engines()
        finally:
            repository.engine.delete_table(tmp_schema, tmp_table)

    def to_data(
        self,
        query: str,
        image: Optional[Union[Image, str]] = None,
        repository: Optional[Repository] = None,
        use_lq: bool = False,
        **kwargs,
    ):
        if image is None:
            if repository is None:
                raise ValueError("repository must be set!")
            # Run the query against the current staging area.
            return self.query_to_data(
                repository.object_engine, query, repository.to_schema(), **kwargs
            )

        # Otherwise, check the image out (full or temporary LQ). Corner case here to fix in the future:
        # if the image is the same as current HEAD and there are no changes, there's no need to do a check out.
        if isinstance(image, str):
            if repository is None:
                raise ValueError("repository must be set!")
            image = repository.images[image]

        if not use_lq:
            image.checkout(force=False)  # Make sure to fail if we have pending changes.
            return self.query_to_data(image.engine, query, image.repository.to_schema(), **kwargs)

        # If we're using LQ, then run the query against a tmp schema
        # (won't download objects unless needed).
        with image.query_schema() as tmp_schema:
            return self.query_to_data(image.engine, query, tmp_schema, **kwargs)


def dedupe_sg_schema(schema_spec: TableSchema, prefix_len: int = 59) -> TableSchema:
    """
    Some foreign schemas have columns that are longer than 63 characters
    where the first 63 characters are the same between several columns
    (e.g. odn.data.socrata.com). This routine renames columns in a schema
    to make sure this can't happen (by giving duplicates a number suffix).
    """

    # We truncate the column name to 59 to leave space for the underscore
    # and 3 digits (max PG identifier is 63 chars)
    prefix_counts: Dict[str, int] = {}
    columns_nums: List[Tuple[str, int]] = []

    for column in schema_spec:
        column_short = column.name[:prefix_len]
        count = prefix_counts.get(column_short, 0)
        columns_nums.append((column_short, count))
        prefix_counts[column_short] = count + 1

    result = []
    for (_, position), column in zip(columns_nums, schema_spec):
        column_short = column.name[:prefix_len]
        count = prefix_counts[column_short]
        if count > 1:
            result.append(
                TableColumn(
                    column.ordinal,
                    f"{column_short}_{position:03d}",
                    column.pg_type,
                    column.is_pk,
                    column.comment,
                )
            )
        else:
            result.append(
                TableColumn(
                    column.ordinal,
                    column.name[:POSTGRES_MAX_IDENTIFIER],
                    column.pg_type,
                    column.is_pk,
                    column.comment,
                )
            )
    return result


def generate_column_names(schema_spec: TableSchema, prefix: str = "col_") -> TableSchema:
    """Replace empty column names with autogenerated ones"""
    result = []
    for i, column in enumerate(schema_spec):
        if column.name:
            result.append(column)
        else:
            result.append(column._replace(name=f"{prefix}{i+1}"))
    return result


def _format_jsonschema(prop, schema, required):
    if prop == "tables":
        return """tables: Tables to mount (default all). If a list, will import only these tables. 
If a dictionary, must have the format
    {"table_name": {"schema": {"col_1": "type_1", ...},
                    "options": {[get passed to CREATE FOREIGN TABLE]}}}."""
    parts = [f"{prop}:"]
    if "description" in schema:
        parts.append(schema["description"])
        if parts[-1][-1] != ".":
            parts[-1] += "."

    if prop in required:
        parts.append("Required.")
    return " ".join(parts)


def build_commandline_help(json_schema):
    required = json_schema.get("required", [])
    return "\n".join(
        _format_jsonschema(p, pd, required) for p, pd in json_schema["properties"].items()
    )


def add_timestamp_tags(repository: Repository, image_hash: str):
    ingestion_time = dt.utcnow()
    short_tag = ingestion_time.strftime("%Y%m%d")
    long_tag = short_tag + "-" + ingestion_time.strftime("%H%M%S")
    new_image = repository.images.by_hash(image_hash)
    new_image.tag(short_tag)
    new_image.tag(long_tag)



SPLITGRAPH_YML_TEMPLATE = """credentials:
  CREDENTIAL_NAME:  # This is the name of this credential that "external" sections can reference.
    plugin: PLUGIN_NAME
    # Credential-specific data matching the plugin's credential schema
    data: {}
repositories:
- namespace: NAMESPACE
  repository: REPOSITORY
  # Catalog-specific metadata for the repository. Optional.
  metadata:
    readme:
      text: Readme
    description: Description of the repository
    topics:
    - sample_topic
  # Data source settings for the repository. Optional.
  external:
    # Name of the credential that the plugin uses. This can also be a credential_id if the
    # credential is already registered on Splitgraph.
    credential: CREDENTIAL_NAME
    plugin: PLUGIN_NAME
    # Plugin-specific parameters matching the plugin's parameters schema
    params: {}
    tables:
      sample_table:
        # Plugin-specific table parameters matching the plugin's schema
        options: {}
        # Schema of the table, a list of objects with `name` and `type`. If set to `[]`, will infer. 
        schema: []
    # Whether live querying is enabled for the plugin (creates a "live" tag in the
    # repository proxying to the data source). The plugin must support live querying.
    is_live: false
    # Ingestion schedule settings. Disable this if you're using GitHub Actions or other methods
    # to trigger ingestion.
    schedule:
"""

DBT_PROJECT_TEMPLATE = """# Sample dbt project referencing data from all ingested/added Splitgraph datasets.
# This is not ready to run, as you'll need to:
#
#   * Manually define tables in your sources (see models/staging/sources.yml, "tables" sections)
#   * Reference the sources using the source(...) macros (see 
#     models/staging/(source_name)/source_name.sql for an example)
#   * Write the actual models

name: 'splitgraph_template'
version: '1.0.0'
config-version: 2

# This setting configures which "profile" dbt uses for this project.
# Note that the Splitgraph runner overrides this at runtime, so this is only useful
# if you are running a local Splitgraph engine and are developing this dbt model against it.
profile: 'splitgraph_template'

target-path: "target"  # directory which will store compiled SQL files
clean-targets:         # directories to be removed by `dbt clean`
  - "target"
  - "dbt_packages"


# Configuring models
# Full documentation: https://docs.getdbt.com/docs/configuring-models

models:
  splitgraph_template:
    # Staging data (materialized as CTEs) that references the source Splitgraph repositories.
    # Here as a starting point. You can reference these models downstream in models that actually
    # materialize as tables.
    staging:
      +materialized: ephemeral
"""

SOURCES_YML_TEMPLATE = """# This file defines all data sources referenced by this model. The mapping
# between the data source name and the Splitgraph repository is in the settings of the dbt plugin
# in splitgraph.yml (see params -> sources)
version: 2
sources:
- name: SOURCE_NAME
  # Splitgraph will use a different temporary schema for this source by patching this project
  # at runtime, so this is for informational purposes only. 
  schema: SCHEMA_NAME
  # We can't currently infer the tables produced by a data source at project generation time,
  # so for now you'll need to define the tables manually.
  tables:
  - name: some_table
"""

SOURCE_TEMPLATE_NOCOM = """name: SOURCE_NAME
schema: SCHEMA_NAME
tables:
- name: some_table
"""

README_TEMPLATE = """# Sample Splitgraph Cloud project

Welcome to the sample Splitgraph Cloud project that we generated for your chosen data sources.

This project contains:

  * [`splitgraph.yml`](./splitgraph.yml): defines live and ingested data sources as well as other
    metadata for your data catalog.
  * [`splitgraph.credentials.yml`](./splitgraph.credentials.yml): defines credentials to your 
    data sources
  * [`build.moveme.yml`](./build.moveme.yml): GitHub Action that runs ingestion / metadata upload
    for all sources:
    * Adds data sources supporting "live" querying (PostgreSQL, MySQL, Elasticsearch, CSV-in-S3) to
      Splitgraph without ingestion, letting you query them at source
    * Runs a "sync" action for other data sources (SaaS etc) to load their data to Splitgraph  
    * Optionally, also runs a dbt project at the end of ingestion to build models.
    
All built repositories are going to be private to your account. You can manage access settings in
the UI by going to https://splitgraph.com/namespace/repository. 

## Required setup

Before you can run this project from GitHub Action, you need to perform a few extra setup steps.

### Add credentials to `splitgraph.credentials.yml`

Edit [`splitgraph.credentials.yml`](./splitgraph.credentials.yml) to add required credentials to
your data sources. **DO NOT COMMIT IT!** You'll add the contents of this file as a secret in the
next step.

### Set up GitHub secrets

Go to the [Secrets page](https://github.com/$GITHUB_REPO/settings/secrets/actions/new) for this
repository and create the following secrets:
  
  * `SPLITGRAPH_CREDENTIALS_YML`: contents of the `splitgraph.credentials.yml` with the data source
    credentals that you've edited in the previous step. 
  * `SPLITGRAPH_API_KEY` / `SPLITGRAPH_API_SECRET`: API keys to Splitgraph Cloud (also known as
    "SQL credentials"). You can get them at https://www.splitgraph.com/settings/sql-credentials (or
    your deployment URL if you're on a private deployment).

### Edit `splitgraph.yml`

We generated a [`splitgraph.yml`](./splitgraph.yml) file from your chosen plugins'
parameters JSONSchema. You should review it and add suitable plugin settings:

  - set `tables` to `tables: {}` to let the plugin automatically infer the schema and the
    options of the data source (by default, it adds a sample table into the project file)
  - change and customize the `metadata` block
  - set up the plugin parameters in `external.params`. Where the comment says `CHOOSE ONE`
    and offers a list of alternative subobjects, choose one entry from the list and delete
    the list itself, leaving the object at the top level.

Example:

```yaml
- namespace: my_namespace
  repository: csv
  # Catalog-specific metadata for the repository. Optional.
  metadata:
    readme:
      text: Readme
    description: Description of the repository
    topics:
    - sample_topic
  # Data source settings for the repository. Optional.
  external:
    # Name of the credential that the plugin uses. This can also be a credential_id if the
    # credential is already registered on Splitgraph.
    credential: csv
    plugin: csv
    # Plugin-specific parameters matching the plugin's parameters schema
    params:
      connection:  # Choose one of:
      - connection_type: http  # REQUIRED. Constant
        url: '' # REQUIRED. HTTP URL to the CSV file
      - connection_type: s3  # REQUIRED. Constant
        s3_endpoint: '' # REQUIRED. S3 endpoint (including port if required)
        s3_bucket: '' # REQUIRED. Bucket the object is in
        s3_region: '' # Region of the S3 bucket
        s3_secure: false # Whether to use HTTPS for S3 access
        s3_object: '' # Limit the import to a single object
        s3_object_prefix: '' # Prefix for object in S3 bucket
      autodetect_header: true # Detect whether the CSV file has a header automatically
      autodetect_dialect: true # Detect the CSV file's dialect (separator, quoting characters etc) automatically
      autodetect_encoding: true # Detect the CSV file's encoding automatically
      autodetect_sample_size: 65536 # Sample size, in bytes, for encoding/dialect/header detection
      schema_inference_rows: 100000 # Number of rows to use for schema inference
      encoding: utf-8 # Encoding of the CSV file
      ignore_decode_errors: false # Ignore errors when decoding the file
      header: true # First line of the CSV file is its header
      delimiter: ',' # Character used to separate fields in the file
      quotechar: '"' # Character used to quote fields
    tables:
      sample_table:
        # Plugin-specific table parameters matching the plugin's schema
        options:
          url: ''  # HTTP URL to the CSV file
          s3_object: '' # S3 object of the CSV file
          autodetect_header: true # Detect whether the CSV file has a header automatically
          autodetect_dialect: true # Detect the CSV file's dialect (separator, quoting characters etc) automatically
          autodetect_encoding: true # Detect the CSV file's encoding automatically
          autodetect_sample_size: 65536 # Sample size, in bytes, for encoding/dialect/header detection
          schema_inference_rows: 100000 # Number of rows to use for schema inference
          encoding: utf-8 # Encoding of the CSV file
          ignore_decode_errors: false # Ignore errors when decoding the file
          header: true # First line of the CSV file is its header
          delimiter: ',' # Character used to separate fields in the file
          quotechar: '"' # Character used to quote fields
        # Schema of the table, a list of objects with `name` and `type`. If set to `[]`, will infer. 
        schema: []
    # Whether live querying is enabled for the plugin (creates a "live" tag in the
    # repository proxying to the data source). The plugin must support live querying.
    is_live: true
    # Ingestion schedule settings. Disable this if you're using GitHub Actions or other methods
    # to trigger ingestion.
    schedule:
```  

becomes:

```yaml
- namespace: my_namespace
  repository: csv
  metadata:
    readme:
      text: Readme
    description: Description of the repository
    topics:
    - sample_topic
  external:
    # No credential required since we're querying a CSV file over HTTP
    plugin: csv
    # Plugin-specific parameters matching the plugin's parameters schema
    params:
      connection:
        connection_type: http  # REQUIRED. Constant
        url: 'https://people.sc.fsu.edu/~jburkardt/data/csv/airtravel.csv' # REQUIRED. HTTP URL to the CSV file
      autodetect_header: true # Detect whether the CSV file has a header automatically
      autodetect_dialect: true # Detect the CSV file's dialect (separator, quoting characters etc) automatically
      autodetect_encoding: true # Detect the CSV file's encoding automatically
      autodetect_sample_size: 65536 # Sample size, in bytes, for encoding/dialect/header detection
      schema_inference_rows: 100000 # Number of rows to use for schema inference
      encoding: utf-8 # Encoding of the CSV file
      ignore_decode_errors: false # Ignore errors when decoding the file
      header: true # First line of the CSV file is its header
      delimiter: ',' # Character used to separate fields in the file
      quotechar: '"' # Character used to quote fields
    # Automatically infer table parameters
    tables: {}
    is_live: true
```

### Set up GitHub Actions

Because this repository was itself generated by a GitHub Actions job, we can't edit the workflow
files for this repository from within the action. You will need to move the job definition file
([`build.moveme.yml`](./build.moveme.yml)) to `.github/workflows/build.yml`.

Optionally, also delete the `seed.yml` file that was used to generate this project.

### Set up dbt and write the models

If you added dbt to this project, this repository also contains a sample dbt project that references
data from all the datasets you've added to it. See [`dbt_project.yml`](./dbt_project.yml) and the
[`models/staging/sources.yml`](models/staging/sources.yml) file for more information.

Currently, we can't infer the columns and the tables that your data sources will produce at this
project generation time, so this dbt project is here as a rough starting point. To get it working,
you will need to: 
 
* Manually define tables in your sources (see 
  [`models/staging/sources.yml`](models/staging/sources.yml), "tables" sections). You might want
  to run the ingestion GitHub Action once first without the dbt step in order to create the
  repositories on Splitgraph and see their tables and columns.
* Write the actual models that reference the sources using the `source(...)` macros (see 
  `models/staging/(source_name)/source_name.sql` for an example)

## Run the action

By default, the generated action waits for a manual trigger to run. You can trigger the action by
going to https://github.com/$GITHUB_REPO/actions/workflows/build.yml and clicking "Run workflow". 

## Next steps
 
  * Edit the GitHub Action to, for example, add a run schedule
  * Browse the ingested and built datasets at https://splitgraph.com/namespace/repository
  * Connect to Splitgraph with an SQL client (see [the docs](https://www.splitgraph.com/docs/splitgraph-cloud/data-delivery-network)) 
"""



"""
Various common functions used by the command line interface.
"""
import itertools
import json
import os
import sys
import time
from functools import wraps
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
from urllib.parse import urlparse

import click
from click.core import Context, Parameter

from splitgraph.cloud.models import JobStatus
from splitgraph.config import REMOTES
from splitgraph.exceptions import RepositoryNotFoundError

if TYPE_CHECKING:
    from splitgraph.core.image import Image
    from splitgraph.core.repository import Repository


class ImageType(click.ParamType):
    """Parser that extracts the full image specification (repository and hash/tag)."""

    name = "Image"

    def __init__(
        self,
        default: Optional[str] = "latest",
        repository_exists: bool = False,
        get_image: bool = False,
    ) -> None:
        """
        :param default: Default tag/hash for image where it's not specified.
        """
        self.default = default
        self.repository_exists = repository_exists
        self.get_image = get_image

    def convert(
        self, value: str, param: Optional[Parameter], ctx: Optional[Context]
    ) -> Tuple["Repository", Optional[Union["Image", str]]]:
        """
        Image specification must have the format [NAMESPACE/]REPOSITORY[:HASH_OR_TAG].

        The parser returns a tuple of (repository object, tag or hash).
        """
        from splitgraph.core.output import parse_repo_tag_or_hash

        repo, tag_or_hash = parse_repo_tag_or_hash(value, default=self.default)

        if self.get_image or self.repository_exists:
            # Check image/repo exists if we're asked (or if we need to produce
            # an actual Image object)
            from splitgraph.core.engine import repository_exists

            if not repository_exists(repo):
                raise RepositoryNotFoundError("Unknown repository %s" % repo)

        if tag_or_hash is not None and self.get_image:
            return repo, repo.images[tag_or_hash]
        else:
            return repo, tag_or_hash


class RepositoryType(click.ParamType):
    name = "Repository"

    def __init__(self, exists: bool = False) -> None:
        self.exists = exists

    def convert(
        self, value: str, param: Optional[Parameter], ctx: Optional[Context]
    ) -> "Repository":
        from splitgraph.core.repository import Repository

        result = Repository.from_schema(value)
        if self.exists:
            from splitgraph.core.engine import repository_exists

            if not repository_exists(result):
                raise RepositoryNotFoundError("Unknown repository %s" % result)
        return result


def load_json_param(value: str, param: Optional[Parameter], ctx: Optional[Context]):
    if value.startswith("@"):
        fopt = click.File(mode="r")
        f = fopt.convert(value[1:], param, ctx)
        try:
            return json.load(f)
        finally:
            f.close()
    return json.loads(value)


class JsonType(click.ParamType):
    """Parser for Json -- a wrapper around json.loads because without specifying
    the name Click shows the type for the option/arg as LOADS.

    Also supports passing JSON files (pass in @filename.json).
    """

    name = "Json"

    def convert(self, value: str, param: Optional[Parameter], ctx: Optional[Context]):
        return load_json_param(value, param, ctx)


def remote_switch_option(*names, **kwargs):
    """
    Adds an option to switch global SG_ENGINE for this invocation of sgr.

    This is useful for e.g. tagging or viewing image information on a remote
    registry. This is not used in operations like commit/checkout (even though
    nothing is preventing SG_ENGINE switch from working on that if the remote engine
    supports this), the user should switch SG_ENGINE envvar themselves in that case.

    :param names: Names
    :param kwargs: Passed to click.option
    """

    if not names:
        names = ("--remote", "-r")

    kwargs.setdefault("default", None)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("help", "Perform operation on a different remote engine")
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("type", click.Choice(REMOTES))

    def switch_engine_back(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            from splitgraph.engine import get_engine, set_engine

            engine = get_engine()
            try:
                f(*args, **kwargs)
                engine.commit()
            except Exception:
                engine.rollback()
                raise
            finally:
                engine.close()

                # In the context of a test run, we need to switch the global engine
                # back to LOCAL (since the engine-switching decorator doesn't
                # get control, so we can't do it there).
                set_engine(get_engine("LOCAL"))

        return wrapped

    def decorator(f):
        def _set_engine(ctx, param, value):
            if not value:
                return
            try:
                from splitgraph.engine import get_engine, set_engine

                engine = get_engine(value)

                set_engine(engine)
            except KeyError:
                raise click.BadParameter("Unknown remote %s!" % value)

        return click.option(*names, callback=_set_engine, **kwargs)(switch_engine_back(f))

    return decorator


def sql_results_to_str(results: List[Tuple[Any]], use_json: bool = False) -> str:
    if use_json:
        import json

        from splitgraph.core.common import coerce_val_to_json

        return json.dumps(coerce_val_to_json(results))

    from tabulate import tabulate

    return tabulate(results, tablefmt="plain")


def emit_sql_results(results, use_json=False, show_all=False):
    if results is None:
        return

    if len(results) > 10 and not show_all:
        click.echo(sql_results_to_str(results[:10], use_json))
        if not json:
            click.echo("...")
    else:
        click.echo(sql_results_to_str(results, use_json))


GQL_POLL_TIME = 5
SPINNER_FREQUENCY = 10


def get_spinner_settings() -> Tuple[Iterator[str], int]:
    from splitgraph.config import SG_CMD_ASCII

    chars = ["|", "/", "-", "\\"] if SG_CMD_ASCII else ["⣾", "⣽", "⣻", "⢿", "⡿", "⣟", "⣯", "⣷"]
    spinner = itertools.cycle(chars)
    poll_interval = max(int(SPINNER_FREQUENCY * GQL_POLL_TIME), 1)
    return spinner, poll_interval


S = TypeVar("S", bound=JobStatus)


def wait_for_job(task_id: str, status_callback: Callable[[], Optional[S]]) -> S:
    """
    Wait for a job to complete (with a CLI spinner) and return its final result.

    :param task_id: ID of the task (used for display)
    :param status_callback: Callback that returns a `JobStatus`
    :return: Final `JobStatus`
    """
    spinner, poll_interval = get_spinner_settings()

    interval = 0
    status_str: Optional[str] = None
    while True:
        if interval % poll_interval == 0:
            status = status_callback()
            if not status:
                raise AssertionError("Job not found")

            if not sys.stdout.isatty() and status_str != status.status:
                click.echo(
                    f" ({status.status or 'PENDING'}) Waiting for task ID {task_id}",
                )
            status_str = status.status

        if sys.stdout.isatty():
            click.echo(f"\033[2K\033[1G{next(spinner)}", nl=False)
            click.echo(f" ({status_str or 'PENDING'}) Waiting for task ID {task_id}", nl=False)
            sys.stdout.flush()
        time.sleep(1.0 / SPINNER_FREQUENCY)
        interval += 1

        if status_str in ("SUCCESS", "FAILURE"):
            click.echo()
            assert status
            return status


def upload_file(file: IO, upload_url: str) -> None:
    """Upload a file to a presigned URL with a tqdm progress bar"""
    import requests
    from tqdm import tqdm
    from tqdm.utils import CallbackIOWrapper

    size = os.fstat(file.fileno()).st_size
    with tqdm(total=size, unit="B", unit_scale=True, unit_divisor=1024) as t:
        wrapped_file = CallbackIOWrapper(t.update, file, "read")
        t.set_description(os.path.basename(file.name))
        requests.put(upload_url, data=wrapped_file)


def download_file(url: str, filename: Optional[str]) -> None:
    """Download a file with a tqdm progress bar"""
    import re

    import requests
    from tqdm import tqdm

    response = requests.get(url, stream=True)
    if not filename and "Content-Disposition" in response.headers:
        filename = re.findall("filename=(.+)", response.headers["Content-Disposition"])[0]
    if not filename:
        filename = urlparse(url).path.split("/")[-1]
    with open(filename, "wb") as f:
        with tqdm.wrapattr(
            f,
            "write",
            miniters=1,
            desc=filename,
            total=int(response.headers.get("content-length", 0)),
        ) as fout:
            for chunk in response.iter_content(chunk_size=4096):
                fout.write(chunk)
    click.echo("Downloaded query results to %s." % filename)



from functools import reduce, wraps
from pathlib import Path
from typing import Callable, Dict, List, Optional, TypeVar

from splitgraph.cloud.project.models import (
    Credential,
    External,
    Metadata,
    Repository,
    SplitgraphYAML,
)
from splitgraph.utils.yaml import safe_dump, safe_load

T = TypeVar("T")


def resolve_optional(
    func: Callable[[T, T], T]
) -> Callable[[Optional[T], Optional[T]], Optional[T]]:
    @wraps(func)
    def wrapped(left: Optional[T], right: Optional[T]) -> Optional[T]:
        if not left:
            return right
        if not right:
            return left
        return func(left, right)

    return wrapped


@resolve_optional
def merge_credentials(
    left: Dict[str, Credential], right: Dict[str, Credential]
) -> Dict[str, Credential]:
    # Credentials override each other fully
    return {**left, **right}


@resolve_optional
def merge_metadata(left: Metadata, right: Metadata) -> Metadata:
    return Metadata.parse_obj(
        {
            **left.dict(by_alias=True, exclude_unset=True),
            **right.dict(by_alias=True, exclude_unset=True),
        }
    )


@resolve_optional
def merge_external(left: External, right: External) -> External:
    left_d = left.dict(by_alias=True, exclude_unset=True)
    right_d = right.dict(by_alias=True, exclude_unset=True)

    for field_name, field_value in right_d.items():
        if field_name in ["credential_id", "credential"]:
            # Credentials (ID or name) override credential references.
            if "credential" in left_d:
                del left_d["credential"]
            if "credential_id" in left_d:
                del left_d["credential_id"]
            left_d[field_name] = field_value
        elif field_name == "tables":
            # Override the schema and the table params separately
            left_d[field_name] = {**left_d.get(field_name, {}), **field_value}
        else:
            left_d[field_name] = field_value
    return External.parse_obj(left_d)


@resolve_optional
def merge_repository(left: Repository, right: Repository) -> Repository:
    return Repository(
        namespace=left.namespace,
        repository=left.repository,
        metadata=merge_metadata(left.metadata, right.metadata),
        external=merge_external(left.external, right.external),
    )


def merge_repository_lists(left: List[Repository], right: List[Repository]) -> List[Repository]:
    left_by_repo = {(r.namespace, r.repository): r for r in left}
    right_by_repo = {(r.namespace, r.repository): r for r in right}

    all_repos = [(r.namespace, r.repository) for r in left]
    all_repos.extend(
        [
            (r.namespace, r.repository)
            for r in right
            if (r.namespace, r.repository) not in left_by_repo
        ]
    )

    result = [
        merge_repository(left_by_repo.get(repo), right_by_repo.get(repo)) for repo in all_repos
    ]
    return [r for r in result if r]


def merge_project_files(left: SplitgraphYAML, right: SplitgraphYAML) -> SplitgraphYAML:
    return SplitgraphYAML(
        credentials=merge_credentials(left.credentials, right.credentials),
        repositories=merge_repository_lists(left.repositories, right.repositories),
    )


def load_project(paths: List[Path]) -> SplitgraphYAML:
    if not paths:
        raise ValueError("No project files specified!")

    def _load(path: Path) -> SplitgraphYAML:
        with open(path, "r") as f:
            return SplitgraphYAML.parse_obj(safe_load(f))

    return reduce(merge_project_files, map(_load, paths))


def dump_project(project: SplitgraphYAML, stream) -> None:
    safe_dump(project.dict(by_alias=True, exclude_unset=True), stream)


def get_source_name(repository: str) -> str:
    return repository.replace("/", "_").replace("-", "_")



from typing import Dict, Union

# Define a ConfigDict: it's a nested dictionary of strings but so far
# we only go down max 2 levels (config["remotes"]["data.splitgraph.com"][param])
# so we enumerate them all explicitly.
ConfigDict = Dict[str, Union[str, Dict[str, str], Dict[str, Dict[str, str]]]]

DEFAULTS: ConfigDict = {
    "SG_ENGINE": "",
    "SG_LOGLEVEL": "WARNING",
    "SG_ENGINE_PREFIX": "splitgraph_engine_",
    "SG_NAMESPACE": "",
    "SG_IS_REGISTRY": "",
    "SG_CHECK_VERSION": "true",
    "SG_ENGINE_FDW_HOST": "localhost",
    "SG_ENGINE_FDW_PORT": "5432",
    "SG_ENGINE_HOST": "localhost",
    "SG_ENGINE_PORT": "5432",
    "SG_ENGINE_DB_NAME": "splitgraph",
    "SG_ENGINE_USER": "sgr",
    "SG_ENGINE_PWD": "supersecure",
    "SG_ENGINE_ADMIN_USER": "",
    "SG_ENGINE_ADMIN_PWD": "",
    "SG_ENGINE_POSTGRES_DB_NAME": "postgres",
    "SG_ENGINE_OBJECT_PATH": "/var/lib/splitgraph/objects",
    # Multicorn/LQFDW currently doesn't give good estimates for GroupBy
    # aggregations (returns 1 distinct group) which makes Postgres use a Sort + GroupAgg
    # aggregation method, so we force it to use hash aggregations for that. This speeds up the
    # US election dataset build by about 30% (82s -> 56s) for the version that uses FROM IMPORT and
    # by about 50% (101s -> 53s) for the version that runs a single big join against multiple images.
    "SG_LQ_TUNING": "SET enable_sort=off; SET enable_hashagg=on;",
    "SG_COMMIT_CHUNK_SIZE": "10000",
    "SG_ENGINE_POOL": "16",
    "SG_CONFIG_FILE": "",
    "SG_META_SCHEMA": "splitgraph_meta",
    "SG_CONFIG_DIRS": "",
    "SG_CONFIG_DIR": "",
    "SG_REPO_LOOKUP": "",
    "SG_REPO_LOOKUP_OVERRIDE": "",
    "SG_S3_HOST": "localhost",
    "SG_S3_PORT": "9000",
    "SG_S3_SECURE": "false",
    "SG_S3_BUCKET": "splitgraph",
    "SG_S3_KEY": "",
    "SG_S3_PWD": "",
    "SG_OBJECT_CACHE_SIZE": "10240",
    "SG_EVICTION_DECAY": "0.002",
    "SG_EVICTION_FLOOR": "1",
    "SG_EVICTION_MIN_FRACTION": "0.05",
    "SG_FDW_CLASS": "splitgraph.core.fdw_checkout.QueryingForeignDataWrapper",
    "SG_CMD_ASCII": "false",
    # Update checks and metrics
    "SG_UPDATE_REMOTE": "data.splitgraph.com",
    "SG_UPDATE_FREQUENCY": "86400",
    "SG_UPDATE_LAST": "0",
    "SG_UPDATE_ANONYMOUS": "false",
    "SG_PLUGIN_DIR": "",
    # Some default sections: these can't be overridden via envvars.
    "external_handlers": {"S3": "splitgraph.hooks.s3.S3ExternalObjectHandler"},
    "data_sources": {
        "postgres_fdw": "splitgraph.hooks.data_source.PostgreSQLDataSource",
        "mongo_fdw": "splitgraph.hooks.data_source.MongoDataSource",
        "mysql_fdw": "splitgraph.hooks.data_source.MySQLDataSource",
        "socrata": "splitgraph.ingestion.socrata.mount.SocrataDataSource",
        "elasticsearch": "splitgraph.hooks.data_source.ElasticSearchDataSource",
        "csv": "splitgraph.ingestion.csv.CSVDataSource",
        "snowflake": "splitgraph.ingestion.snowflake.SnowflakeDataSource",
        "dbt": "splitgraph.ingestion.dbt.data_source.DBTDataSource",
    },
}

ALL_KEYS = list(DEFAULTS.keys())
KEYS = [k for k in ALL_KEYS if k not in ["remotes", "external_handlers", "data_sources"]]
# Keys whose contents we don't print fully
SENSITIVE_KEY_SUFFIXES = ["_PWD", "_TOKEN"]

""" Warning: Every key in DEFAULTS must have a key in ARGUMENT_KEY_MAP
    If you add/remove keys from DEFAULTS, make sure to do so here too.
"""
ARGUMENT_KEY_MAP = {
    "--engine": "SG_ENGINE",
    "--loglevel": "SG_LOGLEVEL",
    "--namespace": "SG_NAMESPACE",
    "--engine-fdw-host": "SG_ENGINE_FDW_HOST",
    "--engine-fdw-port": "SG_ENGINE_FDW_PORT",
    "--engine-host": "SG_ENGINE_HOST",
    "--engine-port": "SG_ENGINE_PORT",
    "--engine-db-name": "SG_ENGINE_DB_NAME",
    "--engine-user": "SG_ENGINE_USER",
    "--engine-pwd": "SG_ENGINE_PWD",
    "--engine-admin-user": "SG_ENGINE_ADMIN_USER",
    "--engine-admin-pwd": "SG_ENGINE_ADMIN_PWD",
    "--engine-postgres-db-name": "SG_ENGINE_POSTGRES_DB_NAME",
    "--engine-object-path": "SG_ENGINE_OBJECT_PATH",
    "--engine-pool": "SG_ENGINE_POOL",
    "--config-file": "SG_CONFIG_FILE",
    "--meta-schema": "SG_META_SCHEMA",
    "--config-dirs": "SG_CONFIG_DIRS",
    "--config-dir": "SG_CONFIG_DIR",
    "--repo-lookup-path": "SG_REPO_LOOKUP",
    "--repo-lookup-override": "SG_REPO_LOOKUP_OVERRIDE",
    "--s3-host": "SG_S3_HOST",
    "--s3-port": "SG_S3_PORT",
    "--s3-access-key": "SG_S3_KEY",
    "--s3-secret-key": "SG_S3_PWD",
    "--s3-bucket": "SG_S3_BUCKET",
    "--object-cache-size": "SG_OBJECT_CACHE_SIZE",
    "--eviction-decay": "SG_EVICTION_DECAY",
    "--eviction-floor": "SG_EVICTION_FLOOR",
    "--eviction-fraction": "SG_EVICTION_MIN_FRACTION",
    "--fdw-class": "SG_FDW_CLASS",
}

ARG_KEYS = list(ARGUMENT_KEY_MAP.keys())

# Reverse of ARGUMENT_KEY_MAP
KEY_ARGUMENT_MAP = {v: k for k, v in ARGUMENT_KEY_MAP.items()}

# ini keys that override environment keys must be same (including SG_)

# Dictionary of keys to a Markdown snippet with documentation. This is compiled
# into a Markdown page referencing the configuration at release time.
# A unit test enforces that each key in KEYS has a documentation in KEY_DOCS.
KEY_DOCS: Dict[str, str] = {
    "SG_ENGINE": """Current engine name in use by Splitgraph. By default, this is the local engine.

This can be overridden to make `sgr` use a different engine in cases where the `--remote` flag is not supported.""",
    "SG_LOGLEVEL": """Logging threshold (log messages not emitted below this).  
Accepted values are CRITICAL, ERROR, WARNING, INFO and DEBUG.  
This can also be changed by passing `--verbosity` to `sgr`, e.g. `sgr --verbosity DEBUG init`.""",
    "SG_ENGINE_PREFIX": "Prefix for Docker containers that are treated as Splitgraph engines by `sgr engine`.",
    "SG_NAMESPACE": "Namespace used by default when pushing to this engine, if not explicitly specified. Normally this is set to the user's username on the registry.",
    "SG_IS_REGISTRY": "Whether this engine is a registry (access only via the SQL API) or an actual Splitgraph engine that supports checkouts.",
    "SG_CHECK_VERSION": "Whether to check the version of the Splitgraph library installed on this engine when a connection to it is first made.",
    "SG_ENGINE_FDW_HOST": "Hostname to use for this engine when it's connecting to itself (during layered querying).",
    "SG_ENGINE_FDW_PORT": "Port to use for this engine when it's connecting to itself (during layered querying).",
    "SG_ENGINE_HOST": "Hostname to use for sgr to connect to the engine.",
    "SG_ENGINE_PORT": "Port to use for sgr to connect to the engine.",
    "SG_ENGINE_DB_NAME": "Database used by Splitgraph.",
    "SG_ENGINE_USER": "Username used by sgr.",
    "SG_ENGINE_PWD": "Password used by sgr.",
    "SG_ENGINE_ADMIN_USER": "Superuser username for the engine, used to first initialize it and create the required Splitgraph schemata and extensions. The one user created on the standard engine is also a superuser, so you only need to override this if you created more than one user on your engine.",
    "SG_ENGINE_ADMIN_PWD": "Superuser password for the engine, used to first initialize it and create the required Splitgraph schemata and extensions. Defaults to the same as SG_ENGINE_PWD.",
    "SG_ENGINE_POSTGRES_DB_NAME": "Name of the default database that the superuser connects to to initialize Splitgraph.",
    "SG_ENGINE_OBJECT_PATH": "Path on the engine's filesystem where Splitgraph physical object files are stored.",
    "SG_LQ_TUNING": "Postgres query planner configuration for Splitfile execution and table imports. This is run before a layered query is executed and allows to tune query planning in case of LQ performance issues. For possible values, see the [PostgreSQL documentation](https://www.postgresql.org/docs/12/runtime-config-query.html).",
    "SG_COMMIT_CHUNK_SIZE": "Default chunk size when `sgr commit` is run. Can be overriden in the command line client by passing `--chunk-size`",
    "SG_ENGINE_POOL": "Size of the connection pool used to download/upload objects. Note that in the case of layered querying with joins on multiple tables, each table will use this many parallel threads to download objects, which can overwhelm the engine. Decrease this value in that case.",
    "SG_CONFIG_FILE": "Location of the Splitgraph configuration file. By default, Splitgraph looks for the configuration in `~/.splitgraph/.sgconfig` and then the current directory.",
    "SG_META_SCHEMA": "Name of the metadata schema. Note that whilst this can be changed, it hasn't been tested and won't be taken into account by engines connecting to this one.",
    "SG_CONFIG_DIRS": "List of directories used to look up the configuration file.",
    "SG_CONFIG_DIR": "Directory the current configuration file is located in.",
    "SG_REPO_LOOKUP": "List of remote names, comma-separated, used for repository lookups during Splitfile execution and `sgr clone` (if a remote name is not specified explicitly).",
    "SG_REPO_LOOKUP_OVERRIDE": "List of overrides for remote engines for some repositories. For example, `override_repo_1:local,override_repo_2:data.splitgraph.com`.",
    "SG_S3_HOST": "Hostname used by the remote engine for object storage. Note that the S3 settings are only used by the remote engine when constructing the URL to give to the client wishing to download/upload objects to S3, not by `sgr` itself.",
    "SG_S3_PORT": "Port used by the remote engine for object storage.",
    "SG_S3_SECURE": "Whether to use HTTPS for object storage.",
    "SG_S3_BUCKET": "S3 bucket used by the engine for object storage.",
    "SG_S3_KEY": "S3 access key.",
    "SG_S3_PWD": "S3 secure key.",
    "SG_OBJECT_CACHE_SIZE": "Object cache size, in megabytes. This only concerns objects downloaded from an external location or a remote engine. When there is no space in the object cache, an eviction is run and objects that haven't been used recently or that are small enough to be easily redownloaded are deleted to free up space.",
    "SG_EVICTION_DECAY": "Significance of recent usage time and object size in cache eviction. See documentation for splitgraph.core.object_manager for an explanation.",
    "SG_EVICTION_FLOOR": "Significance of recent usage time and object size in cache eviction. See documentation for splitgraph.core.object_manager for an explanation.",
    "SG_EVICTION_MIN_FRACTION": "Minimum fraction of the total cache size that has to get freed when an eviction is run. This is to avoid frequent evictions.",
    "SG_FDW_CLASS": "Name of the class used by the layered querying foreign data wrapper on the engine. Internal.",
    "SG_CMD_ASCII": "Set to `true` to disable Unicode output in sgr. Note that `sgr sql` will still output Unicode data.",
    "SG_UPDATE_REMOTE": "Name of the Splitgraph registry to check for sgr updates.",
    "SG_UPDATE_FREQUENCY": "How often to check for updates when sgr is run, in seconds. Set to 0 to disable.",
    "SG_UPDATE_LAST": "Last timestamp an update check was performed. Internal.",
    "SG_UPDATE_ANONYMOUS": "Set to `true` to disable sending the user's ID to the update checker.",
    "SG_PLUGIN_DIR": "Extra directory to look for plugins in. Each subdirectory must have a plugin.py file with a top-level __plugin__ variable pointing at the plugin class",
}



"""Routines for exporting the config back into text."""
from typing import Dict, cast

from splitgraph.config import get_singleton
from splitgraph.config.config import get_all_in_section
from splitgraph.config.keys import DEFAULTS, KEYS, SENSITIVE_KEY_SUFFIXES, ConfigDict


def _kv_to_str(key: str, value: str, no_shielding: bool) -> str:
    value_str = str(value) or ""
    if any(key.endswith(s) for s in SENSITIVE_KEY_SUFFIXES) and not no_shielding:
        value_str = value_str[:1] + "*******"
    return "%s=%s" % (key, value_str)


def serialize_engine_config(
    engine_name: str, conn_params: Dict[str, str], no_shielding: bool
) -> str:
    """
    Output the config section with connection parameters for a single engine.

    :param engine_name: Name of the engine
    :param conn_params: Dictionary of connection parameters
    :param no_shielding: Don't replace passwords with asterisks
    """

    result = "[remote: %s]\n" % engine_name
    result += "\n".join(_kv_to_str(key, value, no_shielding) for key, value in conn_params.items())
    return result


# Parameters that aren't really supposed to be in a config file,
# so we skip them when emitting in the config format.
_SITUATIONAL_PARAMS = ["SG_ENGINE", "SG_CONFIG_FILE"]


def serialize_config(
    config: ConfigDict, config_format: bool, no_shielding: bool, include_defaults: bool = True
) -> str:
    """
    Pretty-print the configuration or print it in the Splitgraph config file format.

    :param config: Configuration dictionary.
    :param config_format: Output configuration in the Splitgraph config file format.
    :param no_shielding: Don't replace sensitive values (like passwords) with asterisks
    :param include_defaults: Emit the config variable even if it's the same as the default.
    :return: Textual representation of the config.
    """

    result = "[defaults]\n" if config_format else ""

    # Emit normal config parameters
    for key in KEYS:
        if config_format and key in _SITUATIONAL_PARAMS:
            continue
        value = get_singleton(config, key)
        if include_defaults or key not in DEFAULTS or value != DEFAULTS[key]:
            result += _kv_to_str(key, value, no_shielding) + "\n"

    # Emit hoisted remotes
    result += "\nCurrent registered remote engines:\n" if not config_format else ""
    for remote, remote_config in get_all_in_section(config, "remotes").items():
        assert isinstance(remote_config, dict)
        if config_format:
            result += "\n" + serialize_engine_config(remote, remote_config, no_shielding) + "\n"
        else:
            result += "\n%s:\n" % remote
            for key, value in remote_config.items():
                result += _kv_to_str(key, value, no_shielding) + "\n"

    # Print Splitfile commands
    if "commands" in config:
        result += "\nSplitfile command plugins:\n" if not config_format else "[commands]\n"
        for command_name, command_class in get_all_in_section(config, "commands").items():
            result += _kv_to_str(command_name, cast(str, command_class), no_shielding) + "\n"

    # Print data sources
    if "data_sources" in config:
        result += "\nData sources:\n" if not config_format else "[data_sources]\n"
        for handler_name, handler_func in get_all_in_section(config, "data_sources").items():
            result += _kv_to_str(handler_name, cast(str, handler_func), no_shielding) + "\n"

    # Print external object handlers
    if "external_handlers" in config:
        result += "\nExternal object handlers:\n" if not config_format else "[external_handlers]\n"
        for handler_name, handler_func in get_all_in_section(config, "external_handlers").items():
            result += _kv_to_str(handler_name, cast(str, handler_func), no_shielding) + "\n"

    return result


def overwrite_config(
    new_config: ConfigDict, config_path: str, include_defaults: bool = False
) -> None:
    """
    Serialize the new config dictionary and overwrite the current config file.
    Note: this will delete all comments in the config!

    :param new_config: Config dictionary.
    :param config_path: Path to the config file.
    :param include_defaults: Whether to include values that are the same
        as their defaults.
    """
    new_config_data = serialize_config(
        new_config, config_format=True, no_shielding=True, include_defaults=include_defaults
    )
    with open(config_path, "w") as f:
        f.write(new_config_data)



import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from psycopg2.sql import SQL, Composable, Composed, Identifier

from splitgraph.config import SPLITGRAPH_API_SCHEMA, SPLITGRAPH_META_SCHEMA
from splitgraph.core.common import adapt, coerce_val_to_json
from splitgraph.core.sql import select
from splitgraph.core.types import Changeset, Comparable, Quals, TableSchema
from splitgraph.engine import ResultShape, validate_type
from splitgraph.engine.postgres.engine import PG_INDEXABLE_TYPES

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PsycopgEngine

T = TypeVar("T", bound=Comparable)


# Custom min/max functions that ignore Nones
def _min(left: Optional[T], right: Optional[T]) -> Optional[T]:
    return right if left is None else (left if right is None else min(left, right))


def _max(left: Optional[T], right: Optional[T]) -> Optional[T]:
    return right if left is None else (left if right is None else max(left, right))


def _inject_collation(qual: str, ctype: str):
    # On engines that have LC_COLLATE set to something like en_US, we still
    # want string comparison to mimic C/Python since we rely on that order
    # in range indexing.
    if ctype in ("text", "varchar", "character varying", "char"):
        return qual + ' COLLATE "C"'
    return qual


def _strip_type_mod(ctype: str) -> str:
    """Convert e.g. numeric(5,3) into numeric for the purpose of
    checking if we can run comparisons on this type and add it to the range index.
    """
    if "(" in ctype:
        ctype = ctype[: ctype.index("(")]
    if "[" in ctype:
        ctype = ctype[: ctype.index("[")]
    return ctype


def _qual_to_index_clause(qual: Tuple[str, str, Any], ctype: str) -> Tuple[SQL, Tuple]:
    """Convert our internal qual format into a WHERE clause that runs against an object's index entry.
    Returns a Postgres clause (as a Composable) and a tuple of arguments to be mogrified into it."""
    column_name, qual_op, value = qual

    # Our index is essentially a bloom filter: it returns True if an object _might_ have rows
    # that affect the result of a query with a given qual and False if it definitely doesn't.
    # Hence, we can combine qualifiers in a similar Boolean way (proof?)

    # If there's no index information for a given column, we have to assume it might match the qual.
    query = SQL("NOT (index -> 'range') ? %s OR ")
    args: List[Any] = [column_name]

    # If the column has to be greater than (or equal to) X, it only might exist in objects
    # whose maximum value is greater than (or equal to) X.
    if qual_op in (">", ">="):
        query += SQL(
            _inject_collation("(index #>> '{{range,{},1}}'", ctype)
            + ")::"
            + ctype
            + "  "
            + qual_op
            + " %s"
        ).format((Identifier(column_name)))
        args.append(value)
    # Similar for smaller than, but here we check that the minimum value is smaller than X.
    elif qual_op in ("<", "<="):
        query += SQL(
            _inject_collation("(index #>> '{{range,{},0}}'", ctype)
            + ")::"
            + ctype
            + " "
            + qual_op
            + " %s"
        ).format((Identifier(column_name)))
        args.append(value)
    elif qual_op == "=":
        query += SQL(
            _inject_collation(
                "%s BETWEEN (index #>> '{{range,{0},0}}')::"
                + ctype
                + " AND (index #>> '{{range,{0},1}}')::"
                + ctype,
                ctype,
            )
        ).format((Identifier(column_name)))
        args.append(value)
    # Currently, we ignore the LIKE (~~) qualifier since we can only make a judgement when the % pattern is at
    # the end of a string.
    # For inequality, we can't really say when an object is definitely not pertinent to a qual:
    #   * if a <> X and X is included in an object's range, the object still might have values that aren't X.
    #   * if X isn't included in an object's range, the object definitely has values that aren't X so we have
    #     to fetch it.
    else:
        # For all other operators, we don't know if they will match so we assume that they will.
        return SQL("TRUE"), ()
    return query, tuple(args)


def _qual_to_sql_clause(qual: Tuple[str, str, str], ctype: str) -> Tuple[Composed, Tuple[str]]:
    """Convert a qual to a normal SQL clause that can be run against the actual object rather than the index."""
    column_name, qual_op, value = qual
    return SQL("{}::" + ctype + " " + qual_op + " %s").format(Identifier(column_name)), (value,)


def _quals_to_clause(
    quals: Optional[Quals],
    column_types: Dict[str, str],
    qual_to_clause: Callable = _qual_to_index_clause,
) -> Tuple[Composable, Tuple]:
    if not quals:
        return SQL(""), ()

    def _internal_quals_to_clause(or_quals):
        clauses, args = zip(
            *[
                qual_to_clause(q, validate_type(_strip_type_mod(column_types[q[0]])))
                for q in or_quals
            ]
        )
        return (
            SQL(" OR ").join(SQL("(") + c + SQL(")") for c in clauses),
            tuple([a for arg in args for a in arg]),
        )

    clauses, args = zip(*[_internal_quals_to_clause(q) for q in quals])
    return (
        SQL(" AND ").join(SQL("(") + c + SQL(")") for c in clauses),
        tuple([a for arg in args for a in arg]),
    )


def quals_to_sql(quals: Optional[Quals], column_types: Dict[str, str]) -> Tuple[Composable, Tuple]:
    """
    Convert a list of qualifiers in CNF to a fragment of a Postgres query
    :param quals: Qualifiers in CNF
    :param column_types: Dictionary of column names and their types
    :return: SQL Composable object and a tuple of arguments to be mogrified into it.
    """

    return _quals_to_clause(quals, column_types, qual_to_clause=_qual_to_sql_clause)


def extract_min_max_pks(
    engine: "PsycopgEngine", fragments: List[str], table_pks: List[str], table_pk_types: List[str]
) -> Any:
    """
    Extract minimum/maximum PK values for given fragments.

    :param engine: Engine the objects live on
    :param fragments: IDs of objects
    :param table_pks: List of columns forming the table primary key
    :param table_pk_types: List of types for table PK columns
    :return: List of min/max primary key for every object.
    """

    # Why can't we use the index for individual columns here? If the PK is composite, consider this example:
    #
    # (1, 1)   CHUNK 1
    # (1, 2)   <- pk1: min 1, max 2; pk2: min1, max2 but (pk1, pk2): min (1, 1), max (2, 1)
    # (2, 1)
    # ------   CHUNK 2
    # (2, 2)   <- pk1: min 2, max 2; pk2: min 2, max 2;
    #
    # Say we have a changeset doing UPDATE pk=(2,2). If we use the index by each part of the key separately,
    # it fits both the first and the second chunk. This essentially means that chunks now overlap,
    # so we'll be fetching/scanning through them when it might not be necessary.

    min_max = []
    pk_sql = SQL(",").join(
        Identifier(p) + SQL(_inject_collation("", t)) for p, t in zip(table_pks, table_pk_types)
    )
    for fragment in fragments:
        query = (
            SQL("SELECT ")
            + pk_sql
            + SQL(" FROM {}.{} ORDER BY ").format(
                Identifier(SPLITGRAPH_META_SCHEMA), Identifier(fragment)
            )
        )
        frag_min = engine.run_sql(
            query + pk_sql + SQL(" LIMIT 1"), return_shape=ResultShape.ONE_MANY
        )
        frag_max = engine.run_sql(
            query
            + SQL(",").join(
                Identifier(p) + SQL(_inject_collation("", t) + " DESC")
                for p, t in zip(table_pks, table_pk_types)
            )
            + SQL(" LIMIT 1"),
            return_shape=ResultShape.ONE_MANY,
        )

        min_max.append((frag_min, frag_max))
    return min_max


def generate_range_index(
    object_engine: "PsycopgEngine",
    object_id: str,
    table_schema: "TableSchema",
    changeset: Optional[Changeset],
    columns: Optional[List[str]] = None,
) -> Dict[str, Tuple[T, T]]:
    """
    Calculate the minimum/maximum values of every column in the object (including deleted values).

    :param object_engine: Engine the object is located on
    :param object_id: ID of the object.
    :param table_schema: Schema of the table
    :param changeset: Changeset (old values will be included in the index)
    :param columns: Columns to run the index on (default all)
    :return: Dictionary of {column: [min, max]}
    """
    columns = columns if columns is not None else [c.name for c in table_schema]

    object_pk = [c.name for c in table_schema if c.is_pk]
    if not object_pk:
        object_pk = [c.name for c in table_schema if c.pg_type in PG_INDEXABLE_TYPES]
    column_types = {c.name: validate_type(_strip_type_mod(c.pg_type)) for c in table_schema}
    columns_to_index = [
        c.name
        for c in table_schema
        if _strip_type_mod(c.pg_type) in PG_INDEXABLE_TYPES and (c.is_pk or c.name in columns)
    ]

    logging.debug("Running range index on columns %s", columns_to_index)
    query = SQL("SELECT ") + SQL(",").join(
        SQL(
            _inject_collation("MIN({0}", column_types[c])
            + "), "
            + _inject_collation("MAX({0}", column_types[c])
            + ")"
        ).format(Identifier(c))
        for c in columns_to_index
    )
    query += SQL(" FROM {}.{}").format(Identifier(SPLITGRAPH_META_SCHEMA), Identifier(object_id))
    result = object_engine.run_sql(query, return_shape=ResultShape.ONE_MANY)
    index = {
        col: (cmin, cmax) for col, cmin, cmax in zip(columns_to_index, result[0::2], result[1::2])
    }
    # Also explicitly store the ranges of composite PKs (since they won't be included
    # in the columns list) to be used for faster chunking/querying.
    if len(object_pk) > 1:
        # Add the PK to the same index dict but prefix it with a dollar sign so that
        # it explicitly doesn't clash with any other columns.
        index["$pk"] = extract_min_max_pks(
            object_engine, [object_id], object_pk, [column_types[c] for c in object_pk]
        )[0]
    if changeset:
        # Expand the index ranges to include the old row values in this chunk.
        # Why is this necessary? Say we have a table of (key (PK), value) and a
        # query "value = 42". Say we have 2 objects:
        #
        #   key | value
        #   1   | 42
        #
        #   key | value
        #   1   | 43   (UPDATED)
        #
        # If we don't include the old value that object 2 overwrote in the index, we'll disregard object 2
        # when inspecting the index for that query (since there, "value" only spans [43, 43]) and give the
        # wrong answer (1, 42) even though we should give (1, 43). Similarly with deletes: if the index for
        # an object doesn't say "some of the values spanning this range are deleted in this chunk",
        # we won't fetch the object.
        #
        # See test_lq_qual_filtering for these test cases.

        # We don't need to do this for the PK since the PK is always specified in deletes.

        # For DELETEs, we put NULLs in the non-PK columns; make sure we ignore them here.
        for _, old_row, _ in changeset.values():
            for col, val in old_row.items():
                # Ignore columns that we aren't indexing because they have unsupported types.
                # Also ignore NULL values.
                if col not in columns_to_index or val is None:
                    continue
                # The audit trigger stores the old row values as JSON so only supports strings and floats/ints.
                # Hence, we have to coerce them into the values returned by the index.
                val = adapt(val, column_types[col])
                index[col] = (_min(index[col][0], val), _max(index[col][1], val))
    range_index = {
        k: (coerce_val_to_json(v[0]), coerce_val_to_json(v[1])) for k, v in index.items()
    }
    return range_index


def filter_range_index(
    metadata_engine: "PsycopgEngine",
    object_ids: List[str],
    quals: Any,
    column_types: Dict[str, str],
) -> List[str]:
    clause, args = _quals_to_clause(quals, column_types)
    query = (
        select("get_object_meta", "object_id", table_args="(%s)", schema=SPLITGRAPH_API_SCHEMA)
        + SQL(" WHERE ")
        + clause
    )

    return cast(
        List[str],
        metadata_engine.run_chunked_sql(
            query, [object_ids] + list(args), return_shape=ResultShape.MANY_ONE, chunk_position=0
        ),
    )
